####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

__all__ = ['Extractor']

####################################################################################################

import hashlib
import os
from pathlib import Path

from bs4 import BeautifulSoup

####################################################################################################

class Extractor:

    ##############################################

    def __init__(self, input_: str | Path, output: str | Path) -> None:
        self._input = input_
        self._output = output

        with open(self._input, 'r') as fh:
            html_doc = fh.read()

        parser = 'html.parser'
        # parser = 'lxml'
        soup = BeautifulSoup(html_doc, parser)
        # print(soup.prettify())

        self._digests = {}

        with open(self._output, 'w') as fh:
            self._fh = fh
            self._write_line(f"# Examples extracted from {self._input}")
            # for item in soup.find_all('div', 'lyx_code_item'):
            for tag in soup.find_all(self._selector):
                if tag.name.startswith('h'):
                    self._handle_h(tag)
                else:
                    self._handle_pre(tag)

    ##############################################

    @staticmethod
    def _selector(tag: BeautifulSoup.Tag) -> BeautifulSoup.Tag:
        return tag.name in ('h1', 'h2') or (
            tag.name == 'div' and tag.has_attr('class') and tag.attrs['class'][0] == "lyx_code_item"
        )

    ##############################################

    def _write_line(self, line: str) -> None:
        self._fh.write(line + os.linesep)

    def _write_lines(self, lines: str) -> None:
        self._fh.write(os.linesep.join(lines) + os.linesep)

    ##############################################

    def _handle_h(self, tag: BeautifulSoup.Tag) -> None:
        RULE = '#'*100
        text = []
        for _ in tag.text.splitlines():
            _ = _.strip()
            if _:
                text.append(f"# {_}")
            else:
                text.append("#")
        self._write_line('')
        self._write_line(RULE)
        self._write_line('#')
        self._write_lines(text)
        self._write_line('#')
        # self._write_line(RULE)

    ##############################################

    def _handle_pre(self, tag: BeautifulSoup.Tag) -> None:
        TRIPLE_QUOTE = '"'*3
        for pre in tag.find_all('pre', 'listings'):
            text = pre.text
            if text:
                text = os.linesep.join([line.rstrip() for line in text.splitlines()])
                digest = hashlib.sha1(text.encode('utf-8')).hexdigest()
                digest = digest[:4]
                if digest not in self._digests:
                    self._digests[digest] = text
                elif self._digests[digest] == text:
                    print("duplicated {digest}")
                    continue
                else:
                    print(f"too short digest {digest}")
                    digest += '-too-short'
                lines = [
                    '',
                    '#'*50,
                    f'# @{pre.sourceline}',
                    '',
                    f's{digest} = {TRIPLE_QUOTE}',
                    text,
                    TRIPLE_QUOTE
                ]
                self._write_lines(lines)

####################################################################################################

if __name__ == '__main__':

    import argparse

    argument_parser = argparse.ArgumentParser(description='Extract examples from ngspice HTML manual')

    # .../ngspice-manuals-git/build_html/manual.html
    argument_parser.add_argument(
        'input', metavar="MANUAL.HTML",
        help='html manual path'
    )

    argument_parser.add_argument(
        'output', metavar="OUTPUT",
        help='output'
    )

    args = argument_parser.parse_args()

    Extractor(args.input, args.output)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'join_dict',
    'join_lines',
    'join_list',
    'prefix_lines',
    'remove_multi_space',
]

####################################################################################################

from typing import Any, Iterable
import os

from .unit import str_spice

####################################################################################################

NEWLINE = os.linesep

####################################################################################################

def prefix_lines(items: Iterable[Any], prefix: str = '') -> list[str]:
    return [
        prefix + str(item)
        for item in items
        if item is not None
    ]   # Fixme: and item

####################################################################################################

def join_lines(items: Iterable[Any], prefix: str = '') -> str:
    return NEWLINE.join(prefix_lines(items, prefix))

####################################################################################################

def join_list(items: Iterable[Any]) -> str:
    # return ' '.join([str_spice(item)
    #                  for item in items
    #                  if item is not None and str_spice(item)])
    values = []
    for item in items:
        if item is not None:
            str_value = str_spice(item)
            if str_value:
                values.append(str_value)
    return ' '.join(values)

####################################################################################################

####################################################################################################
#
# Note:
#   PR #136 has non understood changes
#     https://github.com/FabriceSalvaire/InSpice/pull/136/files
#
####################################################################################################

def join_dict(d: dict[str, Any]) -> str:
    # Fixme: remove trailing _ to key ???
    return ' '.join([
        f'{key}={str_spice(value)}'
        for key, value in sorted(d.items())
        if value is not None
    ])

####################################################################################################

def remove_multi_space(txt: str) -> str:
    """Remove multi-space"""
    # Fixme: tab ???
    new_txt = ''
    last_c = None
    for c in txt:
        if c == ' ' and last_c == ' ':
            continue
        new_txt += c
        last_c = c
    return new_txt



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'PythonDumper',
]

####################################################################################################

import os

# from KiCadTools.Schema import KiCadSchema

####################################################################################################

class PythonDumper:

    def generic_wrapper(element):
        def wrapper(self, symbol):
            return self.on_generic(element, symbol)
        return wrapper

    def generic_model_wrapper(element):
        def wrapper(self, symbol):
            return self.on_generic_model(element, symbol)
        return wrapper

    SYMBOL_MAP = {
        'spice-ngspice:0': None,
        'spice-ngspice:C': generic_wrapper('C'),
        'spice-ngspice:CHOKE': None,
        'spice-ngspice:CURRENT_MEASURE': None,
        'spice-ngspice:Csmall': generic_wrapper('C'),
        'spice-ngspice:DIODE':  generic_model_wrapper('D'),
        'spice-ngspice:INDUCTOR': generic_wrapper('L'),
        'spice-ngspice:ISOURCE': generic_wrapper('I'),
        'spice-ngspice:ISRC_ICTL': None,
        'spice-ngspice:ISRC_VCTL': None,
        'spice-ngspice:NMOS': None,
        'spice-ngspice:OPAMP': None,
        'spice-ngspice:PMOS': None,
        'spice-ngspice:QNPN': None,
        'spice-ngspice:QPNP': None,
        'spice-ngspice:R': generic_wrapper('R'),
        'spice-ngspice:Rsmall': generic_wrapper('R'),
        'spice-ngspice:SWITCH': None,
        'spice-ngspice:TOGGLE': None,
        'spice-ngspice:VSOURCE': generic_wrapper('V'),
        'spice-ngspice:VSRC_ICTL': None,
        'spice-ngspice:VSRC_VCTL': None,
        'spice-ngspice:Vsrc': generic_wrapper('V'),
        'spice-ngspice:ZENOR': None,
    }

    ##############################################

    def __init__(self, kicad_schema, use_InSpice_unit=False):

        self._use_InSpice_unit = use_InSpice_unit
        self._code = []

        for symbol in kicad_schema.symbols_by_reference:
            handler = self.SYMBOL_MAP.get(symbol.lib_name, None)
            if handler is not None:
                _ = handler(self, symbol)
                self._code.append(_)

    ##############################################

    def __str__(self):
        return os.linesep.join(self._code)

    ##############################################

    def _pins(self, symbol):
        pins = []
        for pin in symbol.pins:
            _id = pin.net_id.id
            if _id == 0:
                _id = 'circuit.gnd'
            pins.append(_id)
        return pins

    ##############################################

    def _unit_value(self, element, symbol):
        value = symbol.value
        if not self._use_InSpice_unit:
            return value
        power = value[-1]
        # Fixme: complete...
        #  Meg
        if element == 'R':
            unit = 'Ω'
        elif element == 'C':
            unit = 'F'
        if power in 'pnumk':
            value = value[:-1]
            value = f"{value}@u_{power}{unit}"
        return value

    ##############################################

    def _str_args(self, raw_args):
        args = []
        for arg in raw_args:
            if isinstance(arg, str) and ('@' in arg or arg.startswith('circuit.')):
                pass
            elif isinstance(arg, (int, float)):
                arg = str(arg)
            elif isinstance(arg, (str)):
                # arg = "'" + arg + "'"
                arg = '"' + arg + '"'
            args.append(arg)
        return ', '.join(args)

    ##############################################

    def on_generic(self, element, symbol):
        reference = symbol.reference[len(element):]
        value = self._unit_value(element, symbol)
        args = [reference, *self._pins(symbol), value]
        args_str = self._str_args(args)
        return f"circuit.{element}({args_str})"

    ##############################################

    def on_generic_model(self, element, symbol):
        # Fixme: check XD
        reference = symbol.reference[len(element)+1:]
        args = [reference, symbol.value, *self._pins(symbol)]
        args_str = self._str_args(args)
        return f"circuit.{element}({args_str})"



####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice.Doc.ExampleTools import find_libraries
from InSpice import SpiceLibrary, Circuit, Simulator, plot
from InSpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#?# circuit_macros('buck-converter.m4')

circuit = Circuit('Xspice Control Limit Test')

#add a sinwave source
circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, amplitude=10@u_V, frequency=100@u_kHz)
circuit.V('dd', 'vdd', circuit.gnd, 5@u_V)
circuit.V('ss', 'vss', circuit.gnd, -5@u_V)
# a6 in vdd vss out varlimit
circuit.A('cl', 'in', 'vdd', 'vss', 'out', model='varlimit')
# or you can use the following for diffrential inputs. 
# circuit.A('cl', '%vd(in, gnd)', '%vd(vdd, gnd)', '%vd(vss, gnd)', '%vd(out, gnd)', model='varlimit')
# look at the xspice manual for more information about controlling the limit
circuit.model('varlimit', 'climit', in_offset=0.0, gain=1.0, upper_delta=0.0, lower_delta=0.0, limit_range=2, fraction=False)
# print(circuit)

simulator = Simulator.factory()
simulation = simulator.simulation(circuit, temperature=25, nominal_temperature=25)
# add rshunt option to avoid the matrix singularity when using xspice models. a good value is 1e12 which is 1/gmin
simulation.options(rshunt=1e12)
print(simulation)
analysis = simulation.transient(step_time=1@u_us, end_time=20@u_us, start_time=0@u_ms, max_time = 1@u_ns)
figure, ax = plt.subplots(figsize=(20, 10))
time = analysis.time
ax.plot(time * 1e6, analysis['in'], label='in')
ax.plot(time * 1e6, analysis['out'], label='out')
ax.grid()

ax.set_xlabel('t [us]')
ax.set_ylabel('[V]')

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'buck-converter.png')



#r# ==========
#r#  Pickling
#r# ==========

#r# This example explains how to pickle a simulation.

#r# The first method to save simulation data, it is to use the `pickle <https://docs.python.org/3/library/pickle.html>`_
#r# module provided with the Python standad library.  Pickle implements binary protocols for
#r# serializing and de-serializing a Python object structure.

#r# We will first create a circuit and run a simulation for our example.

####################################################################################################

from pathlib import Path
import pickle

import matplotlib.pyplot as plt

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice import plot

from SimulateCircuit import simulate_circuit

####################################################################################################

pickle_file_path = Path('pickled-simulation.data')
run_simulation = not pickle_file_path.exists()

####################################################################################################

if run_simulation:
    analysis = simulate_circuit()

    #r# We can now save, or pickle, our analysis object to a file.
    print(f"Save {pickle_file_path}")
    with open(pickle_file_path, 'wb') as fh:
        pickle.dump(analysis, fh)

if not run_simulation:
    #r# And later, we can reload the data from the file.
    print(f"Save {pickle_file_path}")
    with open(pickle_file_path, 'rb') as fh:
        analysis = pickle.load(fh)
    simulation = analysis.simulation
    print(f"Simulation date: {simulation.simulation_date}")
    print(f"Simulation duration: {simulation.simulation_duration}")
    print(f"Simulator: {simulation.simulator.SIMULATOR} {simulation.simulator_version}")

    #r# The circuit and the simulation parameters was saved.
    print(simulation.circuit)
    print(str(simulation))

    figure, ax = plt.subplots(1, figsize=(20, 10))
    ax.plot(analysis.out)
    ax.grid()
    ax.set_xlabel('t [s]')
    ax.set_ylabel('[V]')
    plt.tight_layout()
    plt.show()

    #f# save_figure('figure', 'persistence.png')

####################################################################################################

#<r

# In details, pickle recursively serialize all data referenced in the analysis object, except for
# the simulator object which is replaced by its name, and recreated when the data are unpickled.
# Pickle can only serialize Python data structures out of the box.  This is not the case for the
# simulation object, e.g. the NgSpice wrapper contains a library reference.

# Despite the Pickle method do the job very simply, it has counterparts for its ability to handle
# API modifications.  By design, pickled data should be serialized and de-serialized by the the same
# API version.  Thus Pickle is more suited for data caching or to transport analysis data between
# two processes.

# To summarise, you can use Pickle if you just want to cache data temporarily.  But you should avoid
# to use it to save a simulation output.

#r>



#r# ======
#r#  HDF5
#r# ======

#r# This example explains how to save a simulation to a HDF5 file.

####################################################################################################

#r# If you want to save a simulation output without the counterparts of Pickle, then a powerful
#r# approach is to use the `HDF5 <https://www.hdfgroup.org/solutions/hdf5>`_ file format.  This format
#r# was designed to save n-dimensional datasets and will guarantee data could be reloaded later by
#r# another software.  But this approach needs more work than just pickle data.

#r# Data must be organised in the HDF5 file.  A good way to do that is to save at least, the Spice
#r# input, the date of the simulation, the name and version of the simulator, and all the waveforms.

####################################################################################################

from pathlib import Path

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice.Export.Hdf5 import Hdf5File

from SimulateCircuit import simulate_circuit

####################################################################################################

analysis = simulate_circuit()

#r# Create a HDF5 file containing the simulation setting and result.
path = Path('simulation.hdf5')
Hdf5File.save(analysis, path)

#r# Dump the HDF5 file content using the command :code:`h5dump simulation.hdf5`



####################################################################################################

import matplotlib.pyplot as plt
import time
####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice.Doc.ExampleTools import find_libraries
from InSpice import SpiceLibrary, Circuit, Simulator, plot
from InSpice.Unit import *

from InSpice.Spice.NgSpice.Shared import NgSpiceShared
ngspice = NgSpiceShared.new_instance()

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#?# circuit_macros('buck-converter.m4')

circuit = Circuit('Buck Converter')
circuit.include(spice_library['genopa1'])
circuit.include(spice_library['1N5822']) # Schottky diode
circuit.include(spice_library['irf150'])

# From Microchip WebSeminars - Buck Converter Design Example

Vin = 12@u_V
Vout = 5@u_V
ratio = Vout / Vin

Iload = 2@u_A
Rload = Vout / (.8 * Iload)

frequency = 400@u_kHz
period = frequency.period
duty_cycle = ratio * period

ripple_current = .3 * Iload # typically 30 %
ripple_voltage = 50@u_mV

print('ratio =', ratio)
print('RLoad =', Rload)
print('period =', period.canonise())
print('duty_cycle =', duty_cycle.canonise())
print('ripple_current =', ripple_current)

#r# .. math:
#r#      U = L \frac{dI}{dt}

L = (Vin - Vout) * duty_cycle / ripple_current
RL = 37@u_mΩ

#r# .. math:
#r#      dV = dI (ESR + \frac{dt}{C} + \frac{ESL}{dt})

ESR = 30@u_mΩ
ESL = 0
Cout = (ripple_current * duty_cycle) / (ripple_voltage - ripple_current * ESR)

ripple_current_in = Iload / 2
ripple_voltage_in = 200@u_mV
ESR_in = 120@u_mΩ
Cin = duty_cycle / (ripple_voltage_in / ripple_current_in - ESR_in)

L = L.canonise()
Cout = Cout.canonise()
Cin = Cin.canonise()

print('L =', L)
print('Cout =', Cout)
print('Cint =', Cin)

circuit.V('in', 'in', circuit.gnd, Vin)
circuit.C('in', 'in', circuit.gnd, Cin)

# Fixme: out drop from 12V to 4V
# circuit.VCS('switch', 'gate', circuit.gnd, 'in', 'source', model='Switch', initial_state='off')
# circuit.PulseVoltageSource('pulse', 'gate', circuit.gnd, 0@u_V, Vin, duty_cycle, period)
# circuit.model('Switch', 'SW', ron=1@u_mΩ, roff=10@u_MΩ)

# Fixme: Vgate => Vout ???
circuit.X('Q', 'irf150', 'in', 'gate', 'source')
# circuit.PulseVoltageSource('pulse', 'gate', 'source', 0@u_V, Vin, duty_cycle, period)
circuit.R('gate', 'gate', 'clock', 1@u_Ω)
circuit.PulseVoltageSource('pulse', 'clock', circuit.gnd, 0@u_V, 2.*Vin, duty_cycle, period)

circuit.X('D', '1N5822', circuit.gnd, 'source')
inductor = circuit.L(1, 'source', 1, L)
# add a series resistor to model the ESR of the inductor. It helps convergence
inductor.pins[0].add_esr(circuit, value = 10@u_mOhm)
inductor.pins[0].add_current_probe(circuit, name = 'inductor_current')

circuit.R('L', 1, 'out', RL)
circuit.C(1, 'out', circuit.gnd, Cout) # , initial_condition=0@u_V
circuit.R('load', 'out', circuit.gnd, Rload)

####################################################################################################

end_time = 500e-6
circ_str = str(circuit)
options = f'.options TEMP = 25C \n'
options += '.options TNOM = 25C \n'
options += '.options NOINIT \n'
options += '.options RSHUNT = 1e12 \n'
# options += '.ic v(opamp_out) = 0\n'
options += f'.tran 1us {end_time} 0 1ns uic\n'
options += '.end'

circ_str += options 

ngspice.load_circuit(circ_str)
print('Loaded circuit:')
# print(ngspice.listing())

live = True
ngspice.run(background=live)
print('Plots:', ngspice.plot_names)
figure, (ax1, ax2) = plt.subplots(2, 1, figsize=(20, 10), sharex=True)
def update_plots(ax1, ax2, sim_time, analysis):
    ax1.clear()
    ax1.plot(sim_time * 1e6, analysis.out, label='Vout [V]')
    ax1.plot(sim_time * 1e6, analysis['source'], label='Vsource [V]')
    ax1.legend(loc='upper right')
    ax1.grid()
    ax1.set_ylabel('[V]')
    ax1.set_xlabel('t [us]')

    ax2.clear()
    ax2.plot(sim_time * 1e6, analysis.branches['vinductor_current'], label='L2 [A]')
    ax2.legend(loc='upper right')
    ax2.set_ylabel('[A]')
    ax2.set_xlabel('t [us]')
    ax2.grid()
    plt.tight_layout()
    plt.draw()
    plt.pause(0.01)

if not live:
    print(ngspice.status())
    plot_data = ngspice.plot(simulation=None, plot_name=ngspice.last_plot)
    analysis = plot_data.to_analysis()
    update_plots(ax1, ax2, analysis.time, analysis)
    plt.show(block=True)
else:
    simulation_done = False
    while not simulation_done:
        time.sleep(0.1)
        ngspice.halt()
        plot_data = ngspice.plot(simulation=None, plot_name=ngspice.last_plot)
        print(ngspice.status())

        analysis = plot_data.to_analysis()
        sim_time = analysis.time
        if sim_time[-1]._value < end_time - 1e-6:
            ngspice.resume()
        else:
            simulation_done = True

        update_plots(ax1, ax2, sim_time, analysis)
        if simulation_done:
            plt.show(block=True)
            break



#r# =======
#r#  Skidl
#r# =======

# Fixme: elem[a, b, c]

#r# This example explores Skidl...

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

from pprint import pprint

####################################################################################################

from InSpice import Circuit, SubCircuitFactory
from InSpice.Unit import *

circuit = Circuit('Test')

# Capacitor > DipoleElement

print("Actual API")
C1 = circuit.C(1, 1, circuit.gnd, 2@u_uF)
print(C1.plus)
# Pin plus of C1 on node 3
print(type(C1.plus))
# <class 'InSpice.Spice.Netlist.Pin'>
print(type(C1.plus.node))
# <class 'InSpice.Spice.Netlist.Node'>
print(C1.pins)
# [Pin plus of C1 on node 3, Pin minus of C1 on node 4]
print(C1.nodes)
# [Node 3, Node 4]

print()
print("Try dangling...")
# InSpice is happy with that...
C2 = circuit.C(2, None, None, 2@u_uF)
# But in fact, it creates a node 'None'
print(C2.minus)
print(C2.minus.dangling)
# Pin minus of C2 on node None
# dangling impl: now it fails
# print(circuit)
# C2 None None 2uF

print()
print("Connect pins...")
# pin.node is getter only
# it works...
C2.plus.connect(circuit.node(1))
print(type(circuit.gnd), circuit.gnd)
C2.minus.connect(circuit.gnd)
print(C2.minus)
print(circuit)

# Element don't implement []
# C2['foo']
# TypeError: 'Capacitor' object is not subscriptable

print()
print("...")

# C > DipoleElement
print(circuit.C.ELEMENT_CLASS)
pprint(circuit.C.ELEMENT_CLASS.__mro__)
# pprint(circuit.C.ELEMENT_CLASS.__dict__)
print(circuit.C.ELEMENT_CLASS.PINS)
print('C3...')
C3 = circuit.C(3, *[None]*len(circuit.C.ELEMENT_CLASS.PINS), 2@u_uF)
# C3 = circuit.C(3, None, None, 2@u_uF)

print('Connect pin to node...')
print(C3.PINS)
print(C3.PIN_NAMES)
# C3.minus.connect(circuit.gnd)
C3.minus += circuit.gnd
print(C3.minus)
node = circuit.get_node(2, create=True)
C3.plus += node
print(circuit)

in_ = circuit.get_node('in', create=True)
out = circuit.get_node('out', create=True)
R1 = circuit.R(1, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)
R2 = circuit.R(2, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)
R3 = circuit.R(3, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)

# node <= pin
in_ += R1.minus
# node <= pins
out += R1.plus, R2.plus
# circuit.gnd += R2.minus   # want a setter... !
gnd = circuit.gnd
gnd += R2.minus
# dangling <= pin
R3.plus += R2.plus
R3.minus += gnd
print(circuit)

R11 = circuit.R(11, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)
R12 = circuit.R(12, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)
R13 = circuit.R(13, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)

# Create new nodes
R11.plus += R12.plus
R11.minus += circuit.gnd
# pin <= dangling
R11.minus += R13.plus
# merge node
R11.plus += R2.plus

R21 = circuit.R(21, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)
R22 = circuit.R(22, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)
R23 = circuit.R(23, *[None]*len(circuit.R.ELEMENT_CLASS.PINS), 100@u_Ω)

gnd & ( R21 | R22 ) & R23 & out

####################################################################################################

# InSpice don't have "default circuit"
# Need a way to create dangling element

####################################################################################################
#
# Skidl API
#

# The += operator is the only way to make connections!

# Net-to-Net: Connecting one net to another merges the pins on both nets into a single, larger net.
# Pin-to-Net: A pin is connected to a net, adding it to the list of pins connected to that net. If
#             the pin is already attached to other nets, then those nets are merged with this net.
# Net-to-Pin: This is the same as doing a pin-to-net connection.
# Pin-to-Pin: A net is created and both pins are attached to it. If one or both pins are already
#             connected to other nets, then those nets are merged with the newly-created.

#   node1 += node2
#   el[1] += node
#   node += el1[1], el2[1], ...
#   el[1] += el[2]

# One-to-One:   This is the most frequent type of connection, for example, connecting one pin to
#               another or connecting a pin to a net.
# One-to-Many:  This mainly occurs when multiple pins are connected to the same net, like when
#               multiple ground pins of a chip are connected to the circuit ground net.
# Many-to-Many: This usually involves bus connections to a part, such as connecting a bus to the
#               data or address pins of a processor. For this variant, there must be the same number
#               of things to connect in each set, e.g. you can’t connect three pins to four nets.

# += el[1, 2, 3]
# el[1, 2, 3] +=

# Making Serial, Parallel, and Tee Networks

# https://xess.com/skidl/docs/_site/blog/sweetening-skidl

# ser_ntwk = vcc & r1 & r2 & r3 & r4 & gnd
# par_ntwk = vcc & (r1 | r2 | r3 | r4) & gnd
# combo_ntwk = vcc & ((r1 & r2) | (r3 & r4)) & gnd

# polar_ntwk = vcc & r1 & d1['A,K'] & gnd

# ntwk_ce = vcc & r1 & q1['C,E'] & gnd
# ntwk_b = r2 & q1['B']

# ntwk_ce = vcc & r1 & outp & q1['C,E'] & gnd
# ntwk_b = inp & r2 & q1['B']

# pi_ntwk = inp & tee(cs & gnd) & l & tee(cl & gnd) & outp

# The tee function takes any network as its argument and returns the first node of that network to
# be connected into the higher-level network. The network passed to tee can be arbitrarily complex,
# including any combination of parts, &’s, |’s, and tee’s.



###################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

"""This modules implements classes to perform simulations.
"""

####################################################################################################

import logging

####################################################################################################

from ..Simulator import Simulator
from .Server import XyceServer
from .Simulation import XyceSimulation

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class XyceSimulator(Simulator):

    _logger = _module_logger.getChild('XyceSimulator')

    SIMULATOR = 'xyce'

    ##############################################

    def __init__(self, **kwargs):
        # super().__init__(**kwargs)
        xyce_command = kwargs.get('xyce_command', None)
        self._xyce_server = XyceServer(xyce_command=xyce_command)

    ##############################################

    @property
    def version(self):
        # Fixme: How to implement ?
        return ''

    ##############################################

    def simulation(self, circuit, **kwargs):
        return XyceSimulation(self, circuit, **kwargs)

    ##############################################

    def run(self, simulation, *args, **kwargs):

        # Fixme: NEED TO BE TESTED !!!
        raw_file = self._xyce_server(spice_input=str(self))
        raw_file.simulation = simulation

        # for field in raw_file.variables:
        #     print field

        return raw_file.to_analysis()



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'BuiltinFunctions',
    'DotCommands',
    'ElementLetters',
    'UnitSuffix',
]

####################################################################################################

class DotCommands:
    # Fixme: from Ngspice flavour
    AC = "start an ac simulation"
    CONTROL = "start a .control section"
    CSPARAM = "define parameter(s) made available in a control section"
    DC = "start a dc simulation"
    DISTO = "start a distortion analysis simulation"
    ELSE = "conditional branching in the netlist"
    ELSEIF = "conditional branching in the netlist"
    END = "end of the netlist"
    ENDC = "end of the .control section"
    ENDIF = "conditional branching in the netlist"
    ENDS = "end of subcircuit definition"
    FOUR = "Fourier analysis of transient simulation output"
    FUNC = "define a function"
    GLOBAL = "define global nodes"
    IC = "set initial conditions"
    IF = "conditional branching in the netlist"
    INCLUDE = "include part of the netlist"
    LIB = "include a library"
    MEAS = "measurements during the simulation"
    MODEL = "list of device model parameters"
    NODESET = "set initial conditions"
    NOISE = "start a noise simulation"
    OP = "start an operating point simulation"
    OPTIONS = "set simulator options"
    PARAM = "define parameter(s)"
    PLOT = "printer plot during batch simulation"
    PRINT = "tabular listing during batch simulation"
    PROBE = "same as .SAVE: name simulation result vectors to be saved"
    PSS = "start a periodic steady state analysis"
    PZ = "start a pole-zero analysis simulation"
    SAVE = "name simulation result vectors to be saved"
    SENS = "start a sensitivity analysis"
    SUBCKT = "start of subcircuit definitions"
    TEMP = "set the ciruit temperature"
    TF = "start a transfer function analysis"
    TITLE = "title of the netlist"
    TRAN = "start a transient simulation"
    WIDTH = "width of printer plot"

####################################################################################################

class ElementLetters:
    # Fixme: from Ngspice flavour
    # from A to Z
    A = "XSPICE code model"
    B = "Behavioral (arbitrary) source"
    C = "Capacitor"
    D = "Diode"
    E = "Voltage-controlled voltage source (VCVS) linear"
    F = "Current-controlled current source (CCCs) linear"
    G = "Voltage-controlled current source (VCCS) linear"
    H = "Current-controlled voltage source (CCVS) linear"
    I = "Current source"
    J = "Junction field effect transistor (JFET)"
    K = "Coupled (Mutual) Inductors"
    L = "Inductor"
    M = "Metal oxide field effect transistor (MOSFET)"
    N = "Numerical device for GSS"
    O = "Lossy transmission line"
    P = "Coupled multiconductor line (CPL)"
    Q = "Bipolar junction transistor (BJT)"
    R = "Resistor"
    S = "Switch (voltage-controlled)"
    T = "Lossless transmission line"
    U = "Uniformly distributed RC line"
    V = "Voltage source"
    W = "Switch (current-controlled)"
    X = "Subcircuit"
    Y = "Single lossy transmission line (TXL)"
    Z = "Metal semiconductor field effect transistor (MESFET)"

####################################################################################################

class BuiltinFunctions:
    sqrt = ('x')
    sin = ('x')
    cos = ('x')
    tan = ('x')
    sinh = ('x')
    cosh = ('x')
    tanh = ('x')
    asin = ('x')
    acos = ('x')
    atan = ('x')
    asinh = ('x')
    acosh = ('x')
    atanh = ('x')
    arctan = ('x')   # atan = ('x'), kept for compatibility
    exp = ('x')
    ln = ('x')
    log = ('x')
    abs = ('x')
    nint = ('x')       # Nearest integer, half integers towards even
    int = ('x')        # Nearest integer rounded towards 0
    floor = ('x')      # Nearest integer rounded towards -∞
    ceil = ('x')       # Nearest integer rounded towards +∞
    pow = ('x', 'y')   # x raised to the power of y (pow from C runtime library)
    pwr = ('x', 'y')   # pow(fabs = ('x'), y)
    min = ('x', 'y')
    max = ('x', 'y')
    sgn = ('x')   # 1.0 for x > 0, 0.0 for x == 0, -1.0 for x < 0
    ternary_fcn = ('x', 'y', 'z')      # x ? y : z
    gauss = ('nom', 'rvar', 'sigma')   # nominal value plus variation drawn from Gaussian
                                       # distribution with mean 0 and standard deviation rvar
                                       # (relative to nominal), divided by sigma
    agauss = ('nom', 'avar', 'sigma')  # nominal value plus variation drawn from Gaussian
                                       # distribution with mean 0 and standard deviation avar
                                       # (absolute), divided by sigma
    unif = ('nom', 'rvar')    # nominal value plus relative variation (to nominal) uniformly distributed between +/-rvar
    aunif = ('nom', 'avar')   # nominal value plus absolute variation uniformly distributed between +/-avar
    limit = ('nom', 'avar')   # nominal value +/-avar, depending on random number in [-1, 1[ being > 0 or < 0

####################################################################################################

class UnitSuffix:
    T = 'Tera'    # 10**12
    G = 'Giga'    # 10**9
    Meg = 'Mega'  # 10**6
    K = 'Kilo'    # 10**3
    mil = 'Mil'   # 25.4 ×10**−6
    m = 'milli'   # 10**−3
    u = 'micro'   # 10**−6
    n = 'nano'    # 10**−9
    p = 'pico'    # 10**−12
    f = 'femto'   # 10**−15



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import logging
import os

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class DipoleMixin:

    _logger = _module_logger.getChild('DipoleMixin')

    ##############################################

    def __and__(self, other):
        from .Netlist import Node
        self._logger.info(f"Serial connection:{os.linesep}  [{type(self)} {self.name}] & [{type(other)} {other.name}]")
        plus = self.plus
        if isinstance(other, Node):
            plus += other
            return FakeDipole(self.minus, other)
        else:
            plus += other.minus
            return FakeDipole(self.minus, other.plus)

    ##############################################

    def __or__(self, other):
        self._logger.info(f"Parallel connection:{os.linesep}  [{type(self)} {self.name}] | [{type(other)} {other.name}]")
        minus = self.minus
        plus = self.plus
        minus += other.minus
        plus += other.plus
        return FakeDipole(self.minus, self.plus)

    ##############################################

    def __rand__(self, other):
        self._logger.info(f"Serial connection(rand):{os.linesep}  [{type(other)} {other.name}] & [{type(self)} {self.name}]")
        other += self.minus
        return FakeDipole(other, self.plus)

####################################################################################################

class FakeDipole(DipoleMixin):

    ##############################################

    def __init__(self, minus, plus):
        self._minus = minus
        self._plus = plus

    ##############################################

    @property
    def name(self):
        return f'[{self._minus}] => [{self._plus}]'

    ##############################################

    @property
    def minus(self):
        return self._minus

    @property
    def plus(self):
        return self._plus



###################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

"""This modules provides classes to handle analysis parameters.

"""

####################################################################################################

__all__ = [
    'ACAnalysisParameters',
    'AcSensitivityAnalysisParameters',
    'DCAnalysisParameters',
    'DcSensitivityAnalysisParameters',
    'DistortionAnalysisParameters',
    'MeasureParameters',
    'NoiseAnalysisParameters',
    'OperatingPointAnalysisParameters',
    'PoleZeroAnalysisParameters',
    'TransferFunctionAnalysisParameters',
    'TransientAnalysisParameters',
]

####################################################################################################

import logging

####################################################################################################

from ..Unit import as_s, as_Hz
from .StringTools import join_list

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class AnalysisParameters:

    """Base class for analysis parameters"""

    _ANALYSIS_NAME = None

    ##############################################

    @property
    def analysis_name(self):
        return self._ANALYSIS_NAME

    ##############################################

    def to_list(self):
        return ()

    ##############################################

    def __str__(self):
        return '.{0.analysis_name} {1}'.format(self, join_list(self.to_list()))

####################################################################################################

class OperatingPointAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for operating point analysis."""

    _ANALYSIS_NAME = 'op'

####################################################################################################

class DcSensitivityAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for DC sensitivity analysis."""

    _ANALYSIS_NAME = 'sens'

    ##############################################

    def __init__(self, output_variable):
        self._output_variable = output_variable

    ##############################################

    @property
    def output_variable(self):
        return self._output_variable

    ##############################################

    def to_list(self):
        return (self._output_variable,)

####################################################################################################

class AcSensitivityAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for AC sensitivity analysis."""

    _ANALYSIS_NAME = 'sens'

    ##############################################

    def __init__(self, output_variable, variation, number_of_points, start_frequency, stop_frequency):

        if variation not in ('dec', 'oct', 'lin'):
            raise ValueError("Incorrect variation type")

        self._output_variable = output_variable
        self._variation = variation
        self._number_of_points = number_of_points
        self._start_frequency = as_Hz(start_frequency)
        self._stop_frequency = as_Hz(stop_frequency)

    ##############################################

    @property
    def output_variable(self):
        return self._output_variable

    @property
    def variation(self):
        return self._variation

    @property
    def number_of_points(self):
        return self._number_of_points

    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    ##############################################

    def to_list(self):
        return (
            self._output_variable,
            self._variation,
            self._number_of_points,
            self._start_frequency,
            self._stop_frequency
        )

####################################################################################################

class DCAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for DC analysis."""

    _ANALYSIS_NAME = 'dc'

    ##############################################

    def __init__(self, **kwargs):

        self._parameters = []
        for variable, value_slice in kwargs.items():
            variable_lower = variable.lower()
            if variable_lower[0] in ('v', 'i', 'r') or variable_lower == 'temp':
                self._parameters += [variable, value_slice.start, value_slice.stop, value_slice.step]
            else:
                raise NameError('Sweep variable must be a voltage/current source, '
                                'a resistor or the circuit temperature')

    ##############################################

    @property
    def parameters(self):
        return self._parameters

    ##############################################

    def to_list(self):
        return self._parameters

####################################################################################################

class ACAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for AC analysis."""

    _ANALYSIS_NAME = 'ac'

    ##############################################

    def __init__(self, variation, number_of_points, start_frequency, stop_frequency):

        # Fixme: use mixin

        if variation not in ('dec', 'oct', 'lin'):
            raise ValueError("Incorrect variation type")

        self._variation = variation
        self._number_of_points = number_of_points
        self._start_frequency = as_Hz(start_frequency)
        self._stop_frequency = as_Hz(stop_frequency)

    ##############################################

    @property
    def variation(self):
        return self._variation

    @property
    def number_of_points(self):
        return self._number_of_points

    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    ##############################################

    def to_list(self):
        return (
            self._variation,
            self._number_of_points,
            self._start_frequency,
            self._stop_frequency
        )

####################################################################################################

class TransientAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for transient analysis."""

    _ANALYSIS_NAME = 'tran'

    ##############################################

    def __init__(self, step_time, end_time, start_time=0, max_time=None, use_initial_condition=False):

        self._step_time = as_s(step_time)
        self._end_time = as_s(end_time)
        self._start_time = as_s(start_time)
        self._max_time = as_s(max_time, none=True)
        self._use_initial_condition = use_initial_condition

    ##############################################

    @property
    def step_time(self):
        return self._step_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def start_time(self):
        return self._start_time

    @property
    def max_time(self):
        return self._max_time

    @property
    def use_initial_condition(self):
        return self._use_initial_condition

    ##############################################

    def to_list(self):
        return (
            self._step_time,
            self._end_time,
            self._start_time,
            self._max_time,
            'uic' if self._use_initial_condition else None,
        )

####################################################################################################

class MeasureParameters(AnalysisParameters):

    """This class defines measurements on analysis.

    """

    _ANALYSIS_NAME = 'meas'

    ##############################################

    def __init__(self, analysis_type, name, *args):

        _analysis_type = str(analysis_type).upper()
        if _analysis_type not in ('AC', 'DC', 'OP', 'TRAN', 'TF', 'NOISE'):
            raise ValueError('Incorrect analysis type {}'.format(analysis_type))

        self._parameters = [_analysis_type, name, *args]

    ##############################################

    @property
    def parameters(self):
        return self._parameters

    ##############################################

    def to_list(self):
        return self._parameters

####################################################################################################

class PoleZeroAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for pole-zero analysis."""

    _ANALYSIS_NAME = 'pz'

    ##############################################

    def __init__(self, node1, node2, node3, node4, tf_type, pz_type):

        self._nodes = (node1, node2, node3, node4)
        self._tf_type = tf_type   # transfert_function
        self._pz_type = pz_type   # pole_zero

    ##############################################

    @property
    def node1(self):
        return self._nodes[0]

    @property
    def node2(self):
        return self._nodes[1]

    def node3(self):
        return self._nodes[2]

    @property
    def node4(self):
        return self._nodes[3]

    @property
    def tf_type(self):
        return self._tf_type

    @property
    def pz_type(self):
        return self._pz_type

    ##############################################

    def to_list(self):
        return list(self._nodes) + [self._tf_type, self._pz_type]

####################################################################################################

class NoiseAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for noise analysis."""

    _ANALYSIS_NAME = 'noise'

    ##############################################

    def __init__(self, output, src, variation, points, start_frequency, stop_frequency, points_per_summary):

        self._output = output
        self._src = src
        self._variation = variation
        self._points = points
        self._start_frequency = start_frequency
        self._stop_frequency = stop_frequency
        self._points_per_summary = points_per_summary

    ##############################################

    @property
    def output(self):
        return self._output

    @property
    def src(self):
        return self._src

    @property
    def variation(self):
        return self._variation

    @property
    def points(self):
        return self._points

    # Fixme: mixin
    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    @property
    def points_per_summary(self):
        return self._points_per_summary

    ##############################################

    def to_list(self):

        parameters = [
            self._output,
            self._src,
            self._variation,
            self._points,
            self._start_frequency,
            self._stop_frequency,
        ]

        if self._points_per_summary:
            parameters.append(self._points_per_summary)

        return parameters

####################################################################################################

class DistortionAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for distortion analysis."""

    _ANALYSIS_NAME = 'disto'

    ##############################################

    def __init__(self, variation, points, start_frequency, stop_frequency, f2overf1):

        self._variation = variation
        self._points = points
        self._start_frequency = start_frequency
        self._stop_frequency = stop_frequency
        self._f2overf1 = f2overf1

    ##############################################

    @property
    def variation(self):
        return self._variation

    @property
    def points(self):
        return self._points

    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    @property
    def f2overf1(self):
        return self._f2overf1

    ##############################################

    def to_list(self):

        parameters = [
            self._variation,
            self._points,
            self._start_frequency,
            self._stop_frequency,
        ]

        if self._f2overf1:
            parameters.append(self._f2overf1)

        return parameters

####################################################################################################

class TransferFunctionAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for transfer function (.tf) analysis."""

    _ANALYSIS_NAME = 'tf'

    ##############################################

    def __init__(self, outvar, insrc):
        self._outvar = outvar
        self._insrc = insrc

    ##############################################

    @property
    def outvar(self):
        return self._outvar

    @property
    def insrc(self):
        return self._insrc

    ##############################################

    def to_list(self):
        return (self._outvar, self._insrc)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import argparse

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice.Spice.Parser import SpiceFile
from InSpice.Spice.Parser.Translator import Builder, ToPython

####################################################################################################

def main():

    parser = argparse.ArgumentParser(description='Convert a circuit file to InSpice')

    parser.add_argument('circuit_file', # metavar='circuit_file',
                        help='.cir file')

    parser.add_argument('-o', '--output',
                        default=None,
                        help='Output file')

    parser.add_argument('--ground',
                        type=int,
                        default=0,
                        help='Ground node')

    parser.add_argument('--show',
                        default=False, action='store_true',
                        help='Show circuit')

    parser.add_argument('--format',
                        default=False, action='store_true',
                        help='Format circuit')

    parser.add_argument('--build',
                        default=False, action='store_true',
                        help='Build circuit')

    parser.add_argument('--translate',
                        default=False, action='store_true',
                        help='translate circuit')

    args = parser.parse_args()

    ##############################################

    spice_file = SpiceFile(path=args.circuit_file)

    if args.show:
        print('Title header:', spice_file.title)
        print('Subcircuits:')
        for subcircuit in spice_file.subcircuits:
            print(f'  {subcircuit.name}')
        print('Models:')
        for model in spice_file.models:
            print(f'  {model.name}')

    if args.format:
        print(spice_file.to_spice(comment=True, line_length_max=100))

    if args.build:
        Builder().translate(spice_file)

    if args.translate:
        circuit = ToPython().translate(spice_file, ground=args.ground)
        if args.output is not None:
            with open(args.output, 'w') as f:
                f.write(circuit)
        else:
            print(circuit)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'AnyPinElement',
    'DipoleElement',
    'Element',
    # 'ElementParameterMetaClass',
    'FixedPinElement',
    'NPinElement',
    'OptionalPin',
    'Pin',
    'PinDefinition',
    'TwoPortElement',
]

####################################################################################################

from collections import OrderedDict
import logging

####################################################################################################

from .ElementParameter import (
    ParameterDescriptor,
    PositionalElementParameter,
    FlagParameter, KeyValueParameter,
)
from .FakeDipole import FakeDipole
from .StringTools import join_list

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class PinDefinition:

    """This class defines a pin of an element."""

    ##############################################

    def __init__(self, position, name=None, alias=None, optional=False):
        self._position = position
        self._name = name
        self._alias = alias
        self._optional = optional

    ##############################################

    def clone(self):
        # Fixme: self.__class__ ???
        #  unused in code
        return PinDefinition(self._position, self._name, self._alias, self._optional)

    ##############################################

    def __repr__(self):
        _ = f"Pin #{self._position} {self._name}"
        if self._alias:
            _ += f"or {self._alias}"
        if self._optional:
            _ += " optional"
        return _

    ##############################################

    @property
    def position(self):
        return self._position

    @property
    def name(self):
        return self._name

    @property
    def alias(self):
        return self._alias

    @property
    def optional(self):
        return self._optional

####################################################################################################

class OptionalPin:

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

####################################################################################################

class Pin(PinDefinition):

    """This class implements a pin of an element.

    It stores a reference to the element, the name of the pin and the node.

    """

    _logger = _module_logger.getChild('Pin')

    ##############################################

    def __init__(self, element, pin_definition, node):
        super().__init__(pin_definition.position, pin_definition.name, pin_definition.alias)
        self._element = element
        self._node = node
        if self.connected:
            node.connect(self)

    ##############################################

    @property
    def element(self):
        return self._element

    @property
    def node(self):
        return self._node

    ##############################################

    @property
    def dangling(self):
        return self._node is None

    @property
    def connected(self):
        return self._node is not None

    # def __bool__(self):
    #     return self._node is not None

    ##############################################

    def __repr__(self):
        _ = f"Pin {self._name} of {self._element.name}"
        if self.dangling:
            return f"{_} is dangling"
        else:
            return f"{_} on node {self._node}"

    ##############################################

    def connect(self, node):
        # Fixme: if not isinstance(node, Node) ???
        # in ctor netlist.get_node(node, True)
        if self.connected:
            self.disconnect()
        node.connect(self)
        self._node = node

    def disconnect(self):
        # used in Element.detach
        if self.connected:
            self._node.disconnect(self)
            self._node = None

    def __iadd__(self, obj):
        """Connect a node or a pin to the node."""
        from .Netlist import Node
        if isinstance(obj, Node):
            # pin <= node  ===  node <= pin
            self.connect(obj)
        elif isinstance(obj, Pin):
            # pin <=> pin
            if self.connected:
                if obj.connected:
                    # connected <=> connected
                    self._node.merge(obj._node)
                else:
                    # connected <=> dangling
                    obj.connect(self._node)
            else:
                if obj.connected:
                    # dangling <=> connected
                    self.connect(obj._node)
                else:
                    # dangling <=> dangling
                    # Create a new node
                    name = f'{self._element.name}-{self._name}'
                    node = self._element.netlist.get_node(name, create=True)
                    self.connect(node)
                    obj.connect(node)
        else:
            raise ValueError(f"Invalid object {type(obj)}")
        return self

    ##############################################

    def add_current_probe(self, circuit, name=None):
        """Add a current probe between the node and the pin.

        The ammeter is named *ElementName_PinName*.

        """
        # Fixme: require a reference to circuit
        # Fixme: add it to a list
        if self.connected:
            node = self._node
            self._node = '_'.join((self._element.name, str(node), str(self.position) ))
            if name is None:
                name = self._node
            circuit.V(name, node, self._node, '0')
        else:
            raise NameError("Dangling pin")
    
    def add_esr(self, circuit, name= None, value=1e-3):

        """Add a series resistance between the node and the pin.

        The ammeter is named *ElementName_PinName*.

        """

        # Fixme: require a reference to circuit
        # Fixme: add it to a list
        if self.connected:
            node = self._node
            self._node = '_'.join((self._element.name, str(self._node), str(self.position)))
            if name is None:
                name = self._node
            return circuit.R(name, node, self._node, value)
        else:
            raise NameError("Dangling pin")

####################################################################################################

class ElementParameterMetaClass(type):

    # Metaclass to implements the element node and parameter machinery.

    """Metaclass to customise the element classes when they are created and to register SPICE prefix.

    Element classes are of type :class:`ElementParameterMetaClass` instead of :class:`type`

    .. code-block:: none

        class Resistor(metaclass=ElementParameterMetaClass):

        <=>

        Resistor = ElementParameterMetaClass('Foo', ...)

    """

    #: Dictionary for SPICE prefix -> [cls,]
    _classes = {}

    _logger = _module_logger.getChild('ElementParameterMetaClass')

    ##############################################

    def __new__(meta_cls, class_name, base_classes, namespace):

        # __new__ is called for the creation of a class depending of this metaclass, i.e. at module loading
        # It customises the namespace of the new class

        # Collect positional and optional parameters from class attribute dict
        positional_parameters = {}
        parameters = {}
        for attribute_name, obj in namespace.items():
            if isinstance(obj, ParameterDescriptor):
                obj.attribute_name = attribute_name
                if isinstance(obj, PositionalElementParameter):
                    d = positional_parameters
                elif isinstance(obj, (FlagParameter, KeyValueParameter)):
                    d = parameters
                # else:
                #     raise NotImplementedError
                d[attribute_name] = obj

        # Dictionary for positional parameters : attribute_name -> parameter
        namespace['_positional_parameters'] = OrderedDict(
            sorted(list(positional_parameters.items()), key=lambda t: t[1]))

        # Dictionary for optional parameters
        #   order is not required for SPICE, but for unit test
        namespace['_optional_parameters'] = OrderedDict(
            sorted(list(parameters.items()), key=lambda t: t[0]))

        # Positional parameter array
        namespace['_parameters_from_args'] = [
            parameter
            for parameter in sorted(positional_parameters.values())
            if not parameter.key_parameter]

        # Implement alias for parameters: spice name -> parameter
        namespace['_spice_to_parameters'] = {
            parameter.spice_name: parameter
            for parameter in namespace['_optional_parameters'].values()}
        for parameter in namespace['_spice_to_parameters'].values():
            if (parameter.spice_name in namespace
                and parameter.spice_name != parameter.attribute_name):
                _module_logger.error(f"Spice parameter '{parameter.spice_name}' clash with namespace")

        # Initialise pins

        def make_pin_getter(position):
            def getter(self):
                return self._pins[position]
            return getter

        def make_optional_pin_getter(position):
            def getter(self):
                return self._pins[position] if position < len(self._pins) else None
            return getter

        if 'PINS' in namespace and namespace['PINS'] is not None:
            number_of_optional_pins = 0
            pins = []
            for position, pin_definition in enumerate(namespace['PINS']):
                # ensure pin_definition is a tuple
                if isinstance(pin_definition, OptionalPin):
                    optional = True
                    number_of_optional_pins += 1
                    pin_definition = (pin_definition.name,)
                    pin_getter = make_optional_pin_getter(position)
                else:
                    optional = False
                    pin_getter = make_pin_getter(position)
                if not isinstance(pin_definition, tuple):
                    pin_definition = (pin_definition,)
                for name in pin_definition:
                    # Check for name clash
                    if name in namespace:
                        raise NameError(f"Pin {name} of element {class_name} clashes with another attribute")
                    # Add a pin getter in element class
                    namespace[name] = property(pin_getter)
                # Add pin
                pin = PinDefinition(position, *pin_definition, optional=optional)
                pins.append(pin)
            namespace['PINS'] = pins
            namespace['PIN_NAMES'] = [_.name for _ in pins]
            namespace['__number_of_optional_pins__'] = number_of_optional_pins
        else:
            _module_logger.debug(f"{class_name} don't define a PINS attribute")

        return type.__new__(meta_cls, class_name, base_classes, namespace)

    ##############################################

    def __init__(meta_cls, class_name, base_classes, namespace):

        # __init__ is called after the class is created (__new__)

        type.__init__(meta_cls, class_name, base_classes, namespace)

        # Collect basic element classes
        if 'PREFIX' in namespace:
            prefix = namespace['PREFIX']
            if prefix is not None:
                classes = ElementParameterMetaClass._classes
                if prefix in classes:
                    classes[prefix].append(meta_cls)
                else:
                    classes[prefix] = [meta_cls]

    ##############################################

    # Note: These properties are only available from the class object
    #       e.g. Resistor.number_of_pins or Resistor.__class__.number_of_pins

    @property
    def number_of_pins(cls):
        #! Fixme: many pins ???
        number_of_pins = len(cls.PINS)
        if cls.__number_of_optional_pins__:
            return slice(number_of_pins - cls.__number_of_optional_pins__, number_of_pins +1)
        else:
            return number_of_pins

    @property
    def number_of_positional_parameters(cls):
        return len(cls._positional_parameters)

    @property
    def positional_parameters(cls):
        return cls._positional_parameters

    @property
    def optional_parameters(cls):
        return cls._optional_parameters

    @property
    def parameters_from_args(cls):
        return cls._parameters_from_args

    @property
    def spice_to_parameters(cls):
        return cls._spice_to_parameters

####################################################################################################

class Element(metaclass=ElementParameterMetaClass):

    """This class implements a base class for an element.

    It use a metaclass machinery for the declaration of the parameters.

    """

    # These class attributes are defined in subclasses or via the metaclass.
    PINS = None
    _positional_parameters = None
    _optional_parameters = None
    _parameters_from_args = None
    _spice_to_parameters = None

    #: SPICE element prefix
    PREFIX = None

    ##############################################

    def __init__(self, netlist, name, *args, **kwargs):

        self._netlist = netlist
        self._name = str(name)
        self.raw_spice = ''
        self.enabled = True

        # Process remaining args
        if len(self._parameters_from_args) < len(args):
            raise NameError("Number of args mismatch")
        for parameter, value in zip(self._parameters_from_args, args):
            setattr(self, parameter.attribute_name, value)

        # Process kwargs
        for key, value in kwargs.items():
            key = key.lower()
            if key == 'raw_spice':
                self.raw_spice = value
            elif (key in self._positional_parameters or
                  key in self._optional_parameters or
                  key in self._spice_to_parameters):
                setattr(self, key, value)
            elif hasattr(self, 'VALID_KWARGS') and key in self.VALID_KWARGS:
                pass   # cf. NonLinearVoltageSource
            else:
                raise ValueError(f'Unknown argument {key}={value}')

        self._pins = ()
        netlist._add_element(self)

    ##############################################

    def has_parameter(self, name):
        return hasattr(self, '_' + name)

    ##############################################

    def copy_to(self, element):

        for parameter_dict in self._positional_parameters, self._optional_parameters:
            for parameter in parameter_dict.values():
                if hasattr(self, parameter.attribute_name):
                    value = getattr(self, parameter.attribute_name)
                    setattr(element, parameter.attribute_name, value)

        if hasattr(self, 'raw_spice'):
            element.raw_spice = self.raw_spice

    ##############################################

    @property
    def netlist(self):
        return self._netlist

    @property
    def name(self):
        return self.PREFIX + self._name

    @property
    def pins(self):
        return self._pins

    ##############################################

    def detach(self):
        for pin in self._pins:
            pin.disconnect()
        self._netlist._remove_element(self)
        self._netlist = None
        return self

    ##############################################

    @property
    def nodes(self):
        return [pin.node for pin in self._pins]

    @property
    def node_names(self):
        return [str(x) for x in self.nodes]

    ##############################################

    def __repr__(self):
        return self.__class__.__name__ + ' ' + self.name

    ##############################################

    def __setattr__(self, name, value):
        # Implement alias for parameters
        if name in self._spice_to_parameters:
            parameter = self._spice_to_parameters[name]
            object.__setattr__(self, parameter.attribute_name, value)
        elif name in self.PIN_NAMES:
            # __setattr__ is called just after __iadd__
            #   pin += node
            #   means Element.attribute = ( Element.attribute + obj )
            pass
        else:
            object.__setattr__(self, name, value)

    ##############################################

    def __getattr__(self, name):
        # Implement alias for parameters
        if name in self._spice_to_parameters:
            parameter = self._spice_to_parameters[name]
            return object.__getattribute__(self, parameter.attribute_name)
        else:
            raise AttributeError(name)

    ##############################################

    def format_node_names(self):
        """ Return the formatted list of nodes. """
        for pin in self.pins:
            if pin.dangling:
                # yield: NameError: Pin plus of C2 is dangling
                raise NameError(f"{pin}")
        return join_list((self.name, join_list(self.nodes)))

    ##############################################

    def parameter_iterator(self):
        """ This iterator returns the parameter in the right order. """
        # Fixme: .parameters ???
        for parameter_dict in self._positional_parameters, self._optional_parameters:
            for parameter in parameter_dict.values():
                if parameter.nonzero(self):
                    yield parameter

    ##############################################

    # @property
    # def parameters(self):
    #     return self._parameters

    ##############################################

    def format_spice_parameters(self):
        """ Return the formatted list of parameters. """
        return join_list([parameter.to_str(self) for parameter in self.parameter_iterator()])

    ##############################################

    def __str__(self):
        """ Return the SPICE element definition. """
        return join_list((self.format_node_names(), self.format_spice_parameters(), self.raw_spice))

####################################################################################################

class AnyPinElement(Element):

    PINS = ()

    ##############################################

    def copy_to(self, netlist):
        element = self.__class__(netlist, self._name)
        super().copy_to(element)
        return element

####################################################################################################

class FixedPinElement(Element):

    ##############################################

    def __init__(self, netlist, name, *args, **kwargs):

        # Get nodes
        # Usage: if pins are passed using keywords then args must be empty
        #        optional pins are passed as keyword
        pin_definition_nodes = []
        number_of_args = len(args)
        if number_of_args:
            expected_number_of_pins = self.__class__.number_of_pins   # Fixme:
            if isinstance(expected_number_of_pins, slice):
                expected_number_of_pins = expected_number_of_pins.start
            if number_of_args < expected_number_of_pins:
                raise NameError(f"Incomplete node list for element {self.name}")
            else:
                nodes = args[:expected_number_of_pins]
                args = args[expected_number_of_pins:]
                pin_definition_nodes = zip(self.PINS, nodes)
        else:
            for pin_definition in self.PINS:
                if pin_definition.name in kwargs:
                    node = kwargs[pin_definition.name]
                    del kwargs[pin_definition.name]
                elif pin_definition.alias is not None and pin_definition.alias in kwargs:
                    node = kwargs[pin_definition.alias]
                    del kwargs[pin_definition.alias]
                elif pin_definition.optional:
                    continue
                else:
                    raise NameError(f"Node '{pin_definition.name}' is missing for element {self.name}")
                pin_definition_nodes.append((pin_definition, node))

        super().__init__(netlist, name, *args, **kwargs)

        self._pins = [Pin(self, pin_definition, netlist.get_node(node, True))
                      for pin_definition, node in pin_definition_nodes]

    ##############################################

    def copy_to(self, netlist):
        element = self.__class__(netlist, self._name, *self.nodes)
        super().copy_to(element)
        return element

####################################################################################################

class NPinElement(Element):

    PINS = '*'

    ##############################################

    def __init__(self, netlist, name, nodes, *args, **kwargs):
        super().__init__(netlist, name, *args, **kwargs)
        self._pins = [Pin(self, PinDefinition(position), netlist.get_node(node, True))
                      for position, node in enumerate(nodes)]

    ##############################################

    def copy_to(self, netlist):
        nodes = [str(x) for x in self.nodes]
        element = self.__class__(netlist, self._name, nodes)
        super().copy_to(element)
        return element

####################################################################################################

class DipoleElement(FixedPinElement, FakeDipole):
    """This class implements a base class for dipole element."""
    PINS = ('plus', 'minus')

####################################################################################################

class TwoPortElement(FixedPinElement):
    """This class implements a base class for two-port element."""
    PINS = ('output_plus', 'output_minus', 'input_plus', 'input_minus')



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module implements an AST for Spice syntax.
"""

# Fixme: duplicate code

####################################################################################################

__all__ = [
    'Addition',
    'And',
    'AstNode',
    'BinaryOperator',
    'BraceGroup',
    'BracketGroup',
    'Branch',
    'CommaList',
    'Command',
    'Division',
    'DotCommand',
    'Equal',
    'Function',
    'Greater',
    'GreaterEqual',
    'Group',
    'Id',
    'If',
    'InnerParameter',
    'Integer',
    'IntegerDivision',
    'InvertedInput',
    'Less',
    'LessEqual',
    'List',
    'Modulo',
    'Multiplication',
    'Negation',
    'NotEqual',
    'Not',
    'Number',
    'Operator',
    'Or',
    'Set',
    'ParenthesisGroup',
    'PortModifierFunction',
    'PortModifierVector',
    'PortTypeModifier',
    'Power',
    'QuoteGroup',
    'SpaceList',
    'Subtraction',
    'Text',
    'Tuple',
    'UnaryOperator',
]

####################################################################################################

import os

from typing import Optional, Iterator

####################################################################################################

class AstNode:

    ##############################################

    def indentation(self, level: int=0) -> str:
        return ' '*4*level

    ##############################################

    def pretty_print_class(self, level: int=0, linesep: bool=True) -> str:
        _ = self.indentation(level) + self.__class__.__name__
        if linesep:
            _ += os.linesep
        return _

    ##############################################

    @property
    def is_node(self) -> bool:
        return True

    @property
    def is_leaf(self) -> bool:
        return False

    ##############################################

    def __iter__(self) -> Iterator['AstNode']:
        raise NotImplementedError

####################################################################################################

class AstLeaf(AstNode):

    ##############################################

    @property
    def is_node(self) -> bool:
        return False

    @property
    def is_leaf(self) -> bool:
        return True

####################################################################################################
####################################################################################################

class Integer(AstLeaf):

    ##############################################

    def __init__(self, value: int | float) -> None:
        self._value = value

    ##############################################

    def __int__(self) -> int:
        return self._value

    def __float__(self) -> float:
        return self._value

    def __str__(self) -> str:
        return str(self._value)

    ##############################################

    def pretty_print(self, level: int=0) -> str:
        return self.pretty_print_class(level, False) + f' {self._value}'

    ##############################################

    def negate(self) -> 'Number':
        return Number(-self._value)

####################################################################################################

class Number(AstLeaf):

    ##############################################

    def __init__(self, value: int | float, unit: Optional[str]=None, extra_unit: Optional[str]=None) -> None:
        self._value = value
        self._unit = unit
        self._extra_unit = extra_unit
        # self._is_interger = isinstance(value, int) and not unit and not extra_unit

    ##############################################

    @property
    def value(self) -> int | float:
        return self._value

    @property
    def unit(self) -> str:
        return self._unit

    @property
    def extra_unit(self) -> str:
        return self._extra_unit

    @property
    def has_str(self) -> bool:
        return self._unit or self._extra_unit

    ##############################################

    def __float__(self) -> float:
        return float(self._value)

    ##############################################

    def __str__(self) -> str:
        _ = str(self._value)
        if self._unit is not None:
            _ += str(self._unit)
        if self._extra_unit is not None:
            _ += str(self._extra_unit)
        return _

    def pretty_print(self, level: int=0) -> str:
        return self.pretty_print_class(level, False) + f' {self._value} {self._unit} {self._extra_unit}'

    ##############################################

    def negate(self) -> 'Number':
        self._value = - self._value
        return self

####################################################################################################

class Id(AstLeaf):

    ##############################################

    def __init__(self, name: str) -> None:
        self._name = name

    ##############################################

    def __str__(self) -> str:
        return self._name

    ##############################################

    def pretty_print(self, level: int=0) -> str:
        return self.pretty_print_class(level, False) + f' {self._name}'

####################################################################################################

class Text(AstLeaf):

    ##############################################

    def __init__(self, text: str) -> None:
        self._text = text

    ##############################################

    def __str__(self) -> str:
        return self._text

    ##############################################

    def pretty_print(self, level: int=0) -> str:
        return self.pretty_print_class(level, False) + f' {self._text}'

####################################################################################################

class PortTypeModifier(AstLeaf):

    ##############################################

    def __init__(self, name: str) -> None:
        self._name = name

    ##############################################

    @property
    def name(self) -> str:
        return self._name

    ##############################################

    def __str__(self) -> str:
        return self._name

    # def pretty_print(self, level: int=0) -> str:
    #     return self.pretty_print_class(level, False) + f' {self._name}'

####################################################################################################

class Branch(AstLeaf):

    ##############################################

    def __init__(self, name: str) -> None:
        # Fixme: source
        self._name = name

    ##############################################

    def __str__(self) -> str:
        return f'{self._name}#branch'

    def pretty_print(self, level: int=0) -> str:
        return self.pretty_print_class(level, False) + f' {self._name}'

####################################################################################################

class InnerParameter(AstLeaf):

    """
    Format::

        @device_identifier.subcircuit_name.<subcircuit_name_nn>
        +.device_name[parameter]
    """

    ##############################################

    def __init__(self, element_path: str, parameter_name: str) -> None:
        self._element_path = element_path
        self._parameter_name = parameter_name

    ##############################################

    def __str__(self) -> str:
        return f'@{self._element_path}[self._parameter_name]'

    def pretty_print(self, level: int=0) -> str:
        return self.pretty_print_class(level, False) + f' {self._element_path} {self._parameter_name}'

####################################################################################################
####################################################################################################

class InvertedInput(AstNode):

    ##############################################

    def __init__(self, child: str) -> None:
        self._child = child

    ##############################################

    def __iter__(self) -> Iterator[AstNode]:
        yield self._child
        # return iter((self._child,))

    ##############################################

    def __str__(self) -> str:
        return f'!{self._child}'

    def pretty_print(self, level: int=0) -> str:
        txt = self.pretty_print_class(level)
        txt += self._child.pretty_print(level +1)
        return txt

####################################################################################################

class Operator(AstNode):

    NUMBER_OF_OPERANDS = None
    OPERATOR = None

    _unary_operator_map = {}
    _binary_operator_map = {}

    ##############################################

    def __init_subclass__(cls, **kwargs) -> None:
        if cls.OPERATOR is not None:
            if issubclass(cls, BinaryOperator):
                d = cls._binary_operator_map
            elif issubclass(cls, UnaryOperator):
                d = cls._unary_operator_map
            d[cls.OPERATOR] = cls

    ##############################################

    @classmethod
    def get_unary(cls, operator: str):   # -> UnaryOperator:
        return cls._unary_operator_map[operator]

    ##############################################

    @classmethod
    def get_binary(cls, operator: str):   # -> BinaryOperator:
        return cls._binary_operator_map[operator]

    ##############################################

    def __init__(self, *args: AstNode) -> None:
        if (self.NUMBER_OF_OPERANDS is not None
            and len(args) != self.NUMBER_OF_OPERANDS):
            raise ValueError('Wrong number of operands')
        self._childs = args

    ##############################################

    def __iter__(self) -> Iterator[AstNode]:
        return iter(self._childs)

    ##############################################

    def pretty_print(self, level: int=0) -> str:
        indentation = self.indentation(level)
        txt = indentation + self.OPERATOR + os.linesep
        for operand in self:
            txt += operand.pretty_print(level +1).rstrip() + os.linesep
        return txt

####################################################################################################

class UnaryOperator(Operator):

    NUMBER_OF_OPERANDS = 1

    ##############################################

    # def __init__(self, operand1):
    #     super().__init__((operand1,))

    ##############################################

    @property
    def operand(self) -> AstNode:
        return self._childs[0]

    @property
    def operand1(self) -> AstNode:
        return self._childs[0]

    ##############################################

    def __str__(self) -> str:
        return f'{self.OPERATOR} {self.operand}'

    # def pretty_print(self, level: int=0) -> str:
    #     # return ' '.join((self.OPERATOR, str(self.operand1)))
    #     indentation = self.indentation(level)
    #     return indentation + {self.OPERATOR} + ' ' + self.operand1.pretty_print()

####################################################################################################

class BinaryOperator(UnaryOperator):

    NUMBER_OF_OPERANDS = 2

    ##############################################

    # def __init__(self, operand1, operand2):
    #     super().__init__((operand1, operand2))

    ##############################################

    @property
    def operand2(self) -> AstNode:
        return self._childs[1]

    ##############################################

    def __str__(self) -> str:
        return f'{self.operand1} {self.OPERATOR} {self.operand2}'

    # def pretty_print(self, level: int=0) -> str:
    # return ' '.join((str(self.operand1), self.OPERATOR, str(self.operand2)))
    # return f'{self.OPERATOR} {self.operand1} {self.operand2}'

####################################################################################################

# class TernaryOperator(Operator):
#     NUMBER_OF_OPERANDS = 3

####################################################################################################

class Negation(UnaryOperator):
    # fixme: name
    OPERATOR = '-'
    PRECEDENCE = 1

class Not(UnaryOperator):
    OPERATOR = '!'
    PRECEDENCE = 1

####################################################################################################

class Power(BinaryOperator):
    OPERATOR = '**'
    PRECEDENCE = 2

BinaryOperator._binary_operator_map['^'] = Power

class Multiplication(BinaryOperator):
    OPERATOR = '*'
    PRECEDENCE = 3

class Division(BinaryOperator):
    OPERATOR = '/'
    PRECEDENCE = 3

class Modulo(BinaryOperator):
    OPERATOR = '%'
    PRECEDENCE = 3

class IntegerDivision(BinaryOperator):
    OPERATOR = '\\'
    PRECEDENCE = 3

class Addition(BinaryOperator):
    OPERATOR = '+'
    PRECEDENCE = 4

class Subtraction(BinaryOperator):
    OPERATOR = '-'
    PRECEDENCE = 4

####################################################################################################

class Equal(BinaryOperator):
    OPERATOR = '=='
    PRECEDENCE = 5

class NotEqual(BinaryOperator):
    OPERATOR = '!='
    PRECEDENCE = 5

class LessEqual(BinaryOperator):
    OPERATOR = '<='

class GreaterEqual(BinaryOperator):
    OPERATOR = '>='
    PRECEDENCE = 5

class Less(BinaryOperator):
    OPERATOR = '<'
    PRECEDENCE = 5

class Greater(BinaryOperator):
    OPERATOR = '>'
    PRECEDENCE = 5

####################################################################################################

class And(BinaryOperator):
    OPERATOR = '&&'
    PRECEDENCE = 6

class Or(BinaryOperator):
    OPERATOR = '||'
    PRECEDENCE = 7

####################################################################################################

class If(AstNode):

    PRECEDENCE = 8

    ##############################################

    def __init__(self, condition: AstNode, then_expression: AstNode, else_expression: AstNode) -> None:
        self._condition = condition
        self._then_expression = then_expression
        self._else_expression = else_expression

    ##############################################

    @property
    def condition(self) -> AstNode:
        return self._condition

    ##############################################

    @property
    def then_expression(self) -> AstNode:
        return self._then_expression

    ##############################################

    @property
    def else_expression(self) -> AstNode:
        return self._else_expression

    ##############################################

    def __iter__(self) -> Iterator[AstNode]:
        return iter((self._condition, self._then_expression, self._else_expression))

    ##############################################

    def __str__(self):
        return f'{self._condition} ? {self._then_expression} : {self._else_expression}'

    def pretty_print(self, level: int=0) -> str:
        txt = self.pretty_print_class(level)
        txt += self._condition.pretty_print(level +1)
        txt += self._then_expression.pretty_print(level +1).rstrip() + os.linesep
        txt += self._else_expression.pretty_print(level +1).rstrip() + os.linesep
        return txt

####################################################################################################

class Group(AstNode):

    ##############################################

    def __init__(self, child: AstNode) -> None:
        self._child = child

    ##############################################

    def __iter__(self) -> Iterator[AstNode]:
        yield self._child
        # return iter((self._node,))

    ##############################################

    def __str__(self) -> str:
        # if self._child is not None:
        return f'{self.LEFT} {self._child} {self.RIGHT}'

    def pretty_print(self, level: int=0) -> str:
        return self.pretty_print_class(level) + self._child.pretty_print(level +1)

####################################################################################################

class ParenthesisGroup(Group):
    LEFT = '('
    RIGHT = ')'

class BraceGroup(Group):
    LEFT = '{'
    RIGHT = '}'

class QuoteGroup(Group):
    LEFT = "'"
    RIGHT = "'"

class BracketGroup(Group):
    LEFT = '['
    RIGHT = ']'
    # Fixme: CONTEXTUAL SYNTAX !!! in_offset=[0.1 -0.2]

    ##############################################

    def __init__(self, child: AstNode) -> None:
        self._child = self._fix_uminus(child)

    ##############################################

    @classmethod
    def _fix_uminus(cls, child: AstNode) -> AstNode:
        """Fix pattern like::

              -
                  -
                      -
                          Number -0.1 None
                          Number 0.2 None
                      Number 0.3 None
                  Number 0.4 None
        """
        new_child = SpaceList()
        for node in child:
            # Is it a subtraction not within a brace ?
            if isinstance(node, Subtraction):
                for _ in cls._fix_uminus_subtraction(node):
                    new_child.append(_)
            else:
                new_child.append(node)
        return new_child

    ##############################################

    @classmethod
    def _fix_uminus_subtraction(cls, node: Subtraction) -> AstNode:
        left = node.operand1
        right = node.operand2
        if isinstance(left, Subtraction) and isinstance(right, Number):
            # Fix recursive case
            # Replace (- (- ... b) c) by (... -b -c) in the AST
            new_child = SpaceList()
            for _ in cls._fix_uminus_inner(left):
                new_child.append(_)
            new_child.append(right.negate())
        elif isinstance(left, Number) and isinstance(right, Number):
            # Replace (- a b) by (a -b) in the AST
            new_child = SpaceList(left, right.negate())
        else:
            return node
        return new_child

####################################################################################################

class List(AstNode):

    SEPARATOR = ''

    ##############################################

    def __init__(self, *args: AstNode) -> None:
        self._childs = list(args)

    ##############################################

    def append(self, item: AstNode) -> None:
        self._childs.append(item)

    ##############################################

    def __len__(self) -> int:
        return len(self._childs)

    def __bool__(self) -> int:
        return bool(self._childs)

    def __iter__(self) -> Iterator[AstNode]:
        return iter(self._childs)

    def __getitem__(self, _slice: slice) -> AstNode:
        return self._childs[_slice]

    ##############################################

    def __str__(self) -> str:
        return self.SEPARATOR.join([str(_) for _ in self._childs])

    def pretty_print(self, level: int=0) -> str:
        txt = ''
        for obj in self:
            txt += obj.pretty_print(level).rstrip() + os.linesep
        return txt

    ##############################################

    # def iter_on_leaf(self, slice_: slice=None) -> Iterator[tuple[int, AstLeaf]]:
    def iter_on_leaf(self, slice_: slice=None) -> Iterator[AstLeaf]:
        childs = self._childs
        if slice_ is not None:
            childs = childs[slice_]
        for position, child in enumerate(childs):
            if child.is_leaf:
                # yield position, child
                yield child
            else:
                break

    ##############################################

    # def iter_on_set(self, slice_: slice=None) -> Iterator['Set']:
    #     # Fixme: duplicated
    #     childs = self._childs
    #     if slice_ is not None:
    #         childs = childs[slice_]
    #     for child in childs:
    #         if isinstance(child, Set):
    #             yield child

    ##############################################

    def find_first_set(self) -> int:
        for position, child in enumerate(self._childs):
            if isinstance(child, Set):
                return position
        return -1

####################################################################################################

class SpaceList(List):
    SEPARATOR = ' '

class CommaList(List):
    SEPARATOR = ', '

class Tuple(CommaList):

    ##############################################

    def __str__(self) -> str:
        return '(' + super().__str__() + ')'

    def pretty_print(self, level: int=0) -> str:
        txt = self.pretty_print_class(level)
        for obj in self:
            txt += obj.pretty_print(level +1).rstrip() + os.linesep
        return txt

####################################################################################################

class Function(AstNode):

    ##############################################

    def __init__(self, name: str, parameters: CommaList) -> None:
        self._name = name
        self._parameters = parameters

    ##############################################

    def __iter__(self) -> Iterator[AstNode]:
        return iter(self._parameters)

    ##############################################

    @property
    def name(self) -> str:
        return self._name

    ##############################################

    def __str__(self) -> str:
        # if self._parameters:
        return f'{self._name}({self._parameters})'

    def pretty_print(self, level: int=0) -> str:
        txt = self.pretty_print_class(level, False) + f' {self._name}' + os.linesep
        txt += self._parameters.pretty_print(level +1)
        return txt

####################################################################################################

class PortModifierFunction(AstNode):

    ##############################################

    def __init__(self, port_type: PortTypeModifier, parameters: SpaceList) -> None:
        self._port_type = port_type
        self._parameters = parameters

    ##############################################

    def __iter__(self) -> Iterator[AstNode]:
        return iter(self._parameters)

    ##############################################

    def __str__(self) -> str:
        return f'{self._port_type}({self._parameters})'

    def pretty_print(self, level: int=0) -> str:
        txt = self.pretty_print_class(level, False) + f' {self._port_type.name}' + os.linesep
        txt += self._parameters.pretty_print(level +1)
        return txt

####################################################################################################

class PortModifierVector(PortModifierFunction):
    pass

####################################################################################################

class Set(AstNode):

    ##############################################

    def __init__(self, left: AstNode, right: AstNode) -> None:
        self._left = left
        self._right = right

    ##############################################

    def __iter__(self) -> Iterator[AstNode]:
        return iter((self._left, self._right))

    @property
    def left(self) -> AstNode:
        return self._left

    @property
    def right(self) -> AstNode:
        return self._right

    @property
    def left_id(self) -> str:
        if isinstance(self._left, Id):
            return str(self._left)
        else:
            raise ValueError('left is not an Id instance')

    ##############################################

    def __str__(self) -> str:
        return f'{self._left}={self._right}'

    def pretty_print(self, level: int=0) -> str:
        txt = self.pretty_print_class(level)
        txt += self._left.pretty_print(level +1).rstrip() + os.linesep
        txt += self._right.pretty_print(level +1).rstrip() + os.linesep
        return txt

####################################################################################################

class Command(AstNode):

    ##############################################

    def __init__(self, name: str, childs: SpaceList=None) -> None:
        self._name = name
        self._childs = childs

    ##############################################

    def __len__(self) -> int:
        return len(self._childs)

    def __iter__(self) -> Iterator[AstNode]:
        return iter(self._childs)

    def __getitem__(self, _slice: slice) -> Iterator[AstNode]:
        return self._childs[_slice]

    ##############################################

    @property
    def name(self):
        return self._name

    @property
    def childs(self) -> SpaceList:
        return self._childs

    @property
    def first_child(self) -> AstNode:
        return self._childs[0]

    child = first_child

    ##############################################

    def __str__(self) -> str:
        if self._childs:
            return f'{self._name} {self._childs}'
        else:
            return f'{self._name}'

    def pretty_print(self, level: int=0) -> str:
        txt = self.pretty_print_class(level, False) + f' {self._name}' + os.linesep
        if self._childs is not None:
            txt += self._childs.pretty_print(level +1)
        return txt

    ##############################################

    @property
    def first_letter(self) -> str:
        return self._name[0].upper()

    @property
    def after_first_letter(self) -> str:
        return self._name[1:]

    ##############################################

    @property
    def is_dot_command(self) -> bool:
        return self.first_letter == '.'

####################################################################################################

class DotCommand(Command):

    ##############################################

    @property
    def is_dot_command(self) -> bool:
        return True



from InSpice.Doc.ExampleTools import find_libraries
from InSpice import SpiceLibrary, Circuit, Simulator, plot
from InSpice.Unit import *

####################################################################################################

libraries_path = "../skywater-pdk/libraries/sky130_fd_pr/latest/models/sky130.lib.spice"

spice_lib = SpiceLibrary(libraries_path, recurse=True, section="tt")

####################################################################################################

circuit = Circuit('Circuit with Subcircuits')


# # Safely iterate through subcircuits with error handling
# print("\nSubcircuit details:")
# for subcirc in spice_lib.subcircuits:
#     try:
#         print(f"\nProcessing subcircuit: {subcirc}")
        
#         # Safely access the subcircuit
#         subckt = spice_lib[subcirc]
#         print(f"  Successfully loaded subcircuit: {subcirc}")
        
#         # Check what attributes are available
#         if hasattr(subckt, "path"):
#             print(f"  Path: {subckt.path}")
        
#         # Try to access nodes safely
#         if hasattr(subckt, "_nodes"):
#             print(f"  Nodes: {len(subckt._nodes)} nodes found")
#             for i, node in enumerate(subckt._nodes):
#                 print(f"    Node {i+1}: {node}")
#         else:
#             print("  No _nodes attribute found")
            
#         # Check for additional attributes
#         for attr in ["name", "pin_names", "description"]:
#             if hasattr(subckt, attr):
#                 print(f"  {attr}: {getattr(subckt, attr)}")
                
#     except KeyError as e:
#         print(f"  Error: Could not find subcircuit '{subcirc}' - {e}")
#     except Exception as e:
#         print(f"  Error processing subcircuit '{subcirc}': {e}")

# Print information about the library
print(f"Library loaded from: {libraries_path}")
print(f"Available subcircuits in library: {list(spice_lib.subcircuits)}")
print(f"Total number of subcircuits found: {len(list(spice_lib.subcircuits))}")
# Print the circuit
print("\nCircuit:")
print(circuit)


####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

# Fixme: versus InSpice.Plot ???

####################################################################################################

"""This module implements plotting helper."""

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

def plot(waveform, *args, **kwargs):

    """Plot a waveform using the current Axes instance or the one specified by the *axis* key
    argument. Additional parameters are passed to the Matplotlib plot function.

    """

    axis = kwargs.get('axis', plt.gca())
    if 'axis' in kwargs:
        del kwargs['axis']
    axis.plot(waveform.abscissa, waveform, *args, **kwargs)



####################################################################################################

#r#
#r# ========================
#r#  Parse NgSpice Examples
#r# ========================
#r#
#r# This example shows a to use the module :mod:`InSpice.Spice.NgSpice.ManualExamples`.
#r#

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from pathlib import Path
import os

from InSpice.Spice.NgSpice import ManualExamples
from InSpice.Spice.Parser import SpiceSource, SpiceFile

####################################################################################################

classes_to_process = (
    'CircuitDescription',
    'CircuitElementsModels',
    'VoltageCurrentSources',
    'BehavioralSources',
    'TransmissionLines',
    'Diodes',
    'BJT',
    'JFETS',
    'MESFETS',
    'MOSFETS',
    'XSPICE',
    # 'VerilogADeviceModel',
    # 'TCAD',
    'AnalysesOutputControl',
    # 'StartingNgspice',
    # 'InteractiveInterpreter',
    # 'NgspiceUserInterfaces',
    # 'NgspiceSharedLibrary',
    # 'TCLspice',
    'ExampleCircuits1',
    'StatisticalCircuitAnalysis',
    'CircuitOptimization',
    'XspiceBasics',
    # 'ExecutionProcedures',
    'ExampleCircuits2',
    # 'CodeModels',
    # 'ErrorMessages',
    # 'CIDER',
    'ModelDeviceParameters',
)

def generator():
    line_counter = 0
    for cls in ManualExamples.Examples.subclasses():
        if cls.__name__ in classes_to_process:
            for _, value in cls.iter_on_examples(label='E'):
                for line in value.strip().splitlines():
                    # print(f'{line_counter}>>> {line}')
                    if line:
                        yield line_counter, line
                    line_counter += 1

# spice_source = SpiceSource()
# spice_source.read(generator(), title_line=False)

# #r# Dump the pre-processed lines
# for line in spice_source.lines:
#     print(line)

# #r# Note: set env InSpiceLogLevel=debug
# spice_source.parse()

# #r# Dump the parsed lines
# for obj in spice_source.obj_lines:
#     print('-'*80)
#     print(obj.line)
#     print()
#     print(obj.ast.pretty_print())
#     print(obj)

####################################################################################################

spice_examples_path = Path(__file__).parents[1].joinpath('spice-examples')

for filename in (
        'ac-coupled-transistor-amplifier.cir',
        'operational-amplifier-model-1.cir',
):
    print('='*100)
    print(f'File {filename}')
    path = spice_examples_path.joinpath(filename)
    spice_file = SpiceFile(path)



from setuptools import setup, find_packages
import os
import re

# Read the version from __init__.py
with open(os.path.join('InSpice', '__init__.py'), 'r') as f:
    init_content = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", init_content, re.M)
    version = version_match.group(1) if version_match else '0.1.0'

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="InSpice",
    version=version,
    description="Python interface to Ngspice and Xyce circuit simulators (forked from PySpice)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Innovoltive",
    author_email="info@innovoltive.com",
    url="https://github.com/Innovoltive/InSpice",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "matplotlib",
        "PyYAML",
        "cffi>=1.0.0",
        "diskcache",
        "h5py",
    ],
    extras_require={
        "dev": ["pytest", "twine", "build"],
    },
    entry_points={
        "console_scripts": [
            "inspice-post-installation=InSpice.Scripts.InSpice_post_installation:main",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    python_requires='>=3.8',
)


####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2024 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'str_spice',
    'str_spice_list',
]

####################################################################################################

import logging

####################################################################################################

from InSpice.Unit.Unit import UnitValue

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

SPICE_SUFFIX = {
    'Tera': 'T',
    'Giga': 'G',
    'Mega': 'Meg',
    'Kilo': 'K',
    'Mil': 'mil',
    'milli': 'm',
    'micro': 'u',
    'nano': 'n',
    'pico': 'p',
    'femto': 'f',
    'atto': 'a',
}
_ = {}
for name, suffix in SPICE_SUFFIX.items():
    if name[0].isupper():
        _[name.lower()] = suffix
SPICE_SUFFIX.update(_)

try:
    import pint

    @pint.register_unit_format('spice')
    def format_spice_unit(unit, registry, **options) -> str:
        # See Ngspice Manual — 2.1.4.2 Numbers
        # See https://pint.readthedocs.io/en/stable/user/formatting.html#string-formatting-specification
        # registry.parse_unit_name
        #   return tuple of tuples (str, str, str)
        #   all non-equivalent combinations of (prefix, unit name, suffix)
        unit_tuple = registry.parse_unit_name(str(unit), case_sensitive=False)
        prefix = unit_tuple[0][0]
        if prefix and prefix not in SPICE_SUFFIX:
            raise ValueError(f"unsupported Pint unit {unit}")
        _ = str(unit)
        # Fixme: useless ?
        if _.startswith('mega'):
            _ = 'Mega' + _[4:]
        return _
except ImportError:
    pint = None

####################################################################################################

def str_spice(obj, unit=True):
    # Fixme: right place ???
    '''Convert an object to a Spice compatible string.'''
    if isinstance(obj, UnitValue):
        if unit:
            return obj.str_spice()
        else:   # Fixme: ok ???
            return obj.str(spice=False, space=False, unit=False)
    elif pint and isinstance(obj, pint.Quantity):
        # _module_logger.info(f'Pint {obj} -> {obj:spice}')
        return f"{obj:spice}".replace(' ', '')
    else:
        return str(obj)

####################################################################################################

def str_spice_list(*args):
    return [str_spice(x) for x in args]



####################################################################################################

#r# ==================
#r#  Simulation Cache
#r# ==================

# It can be useful to cache the simulations, for example to avoid to rerun a simulation that
# required a long computing time.

# A simulation cache can be implemented by different ways, from the simpler to the complex one.  A
# first idea, is to use a checksum of the SPICE input, e.g. an algorithm like SHA-256, as a key to
# check if the simulation is already available.  Of course, we can check for a match that the two
# inputs are equal.  Then we can simply save the data in a file identified by the checksum.

####################################################################################################

from pathlib import Path

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice.Cache import SimulationCache

from SimulateCircuit import simulate_circuit

####################################################################################################

cache = SimulationCache()
print(cache.directory)

analysis = simulate_circuit()
print(cache.simulation_key(analysis.simulation))

# data
# data
# simulator identity and release



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'is_yaml',
    'SpiceInclude',
]

####################################################################################################

from collections import namedtuple
from datetime import datetime
from pathlib import Path
from typing import Iterator
import hashlib
import logging
import os
import re
import yaml

from InSpice.Spice.Parser import SpiceFile, ParseError
# from InSpice.Spice.Parser import Subcircuit as ParserSubcircuit
# from InSpice.Spice.Parser import Model as ParserModel

####################################################################################################

NEWLINE = os.linesep

_module_logger = logging.getLogger(__name__)

####################################################################################################

YAML_EXTENSION = '.yaml'

def is_yaml(path: Path) -> bool:
    _ = path.suffix.lower()
    return _ == YAML_EXTENSION

####################################################################################################

Pin = namedtuple('Pin', ['index', 'name', 'internal_node'])

####################################################################################################

class Mixin:

    ##############################################

    @classmethod
    def from_yaml(cls, include: 'SpiceInclude', data: dict) -> 'Mixin':
        args = [data[_] for _ in ('name', 'description')]
        return cls(include, *args)

    ##############################################

    def __init__(self, include: 'SpiceInclude', name: str, description: str = '') -> None:
        self._include = include
        self._name = name
        self._description = description

    ##############################################

    @property
    def include(self) -> 'SpiceInclude':
        return self._include

    @property
    def path(self) -> Path:
        return self._include.path

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description

    ##############################################

    def to_yaml(self) -> dict:
        return {
            'name': self._name,
            'description': self._description,
        }

####################################################################################################

class Model(Mixin):

    ##############################################

    @classmethod
    def from_yaml(cls, include: 'SpiceInclude', data: dict) -> 'Model':
        args = [data[_] for _ in ('name', 'type', 'description')]
        return cls(include, *args)

    ##############################################

    def __init__(self, include: 'SpiceInclude', name: str, type_: str, description: str = '') -> None:
        super().__init__(include, name, description)
        self._type = type_

    ##############################################

    @property
    def type(self) -> str:
        return self._type

    ##############################################

    def to_yaml(self) -> dict:
        _ = super().to_yaml()
        _.update({'type': self._type})
        return _

    ##############################################

    def __repr__(self) -> str:
        return f'Model {self._name} {self._type} "{self._description}"'

####################################################################################################

# Fixme: SubCircuit to be consistent with netlist
class Subcircuit(Mixin):

    """
    Subcircuit definitions are for example:
    ```
    .SUBCKT 1N4148 1 2

    .SUBCKT d1n5919brl 2 1

    .SUBCKT LMV981 1 3 6 2 4 5
    # for pinout: +IN -IN +V -V OUT NSD
    ```

    and a call is:
    ```
    X1 2 4 17 3 1 MULTI
    ```
    """

    _logger = _module_logger.getChild('Subcircuit')

    ##############################################

    @classmethod
    def from_yaml(cls, include: 'SpiceInclude', data: dict) -> 'Subcircuit':
        args = [data[_] for _ in ('name', 'nodes', 'description')]
        return cls(include, *args)

    ##############################################

    def __init__(self, include: 'SpiceInclude', name: str, nodes: list[str], description: str = '') -> None:
        super().__init__(include, name, description)
        self._valid = False
        self._nodes = []
        self._pin_map = {}
        self._parse_nodes(nodes)

    ##############################################

    def _parse_nodes(self, nodes: list[str]) -> None:
        self._valid = True
        for index, node_str in enumerate(nodes):
            internal_node = None
            name = None
            if isinstance(node_str, dict):
                # if we are hear we probably have read the yaml file
                internal_node = node_str['internal_node']
                node_str = node_str['name']            
            if isinstance(node_str, str):
                node_str = node_str.strip()
                # make sure the node is a valid spice node identifier
                reg_ex = r'''
                    (?i:                            # case-insensitive
                        (?:[+\-\%](?=[a-z_]))?      # optional +, - or % at the start, only if followed by letter/underscore
                        [a-z0-9_]+                 # one or more letters/digits/underscores
                        (?:\.[a-z0-9_]+)?          # optionally a dot, followed by one or more letters/digits/underscores
                        (?:(?<=[a-z0-9])[+\-]      # optionally a plus or minus if it's right after a letter/digit
                        (?![a-z0-9.])             # and not followed by letter/digit/dot
                        )?
                    )(?!:)                         # negative lookahead: do not allow a colon immediately after
                '''
                # Using VERBOSE to ignore whitespace/comments in the regex
                pattern = re.compile(reg_ex, re.VERBOSE)
                match = pattern.fullmatch(node_str)
                if match:
                    name = match.group(0)
                    if not internal_node: 
                        internal_node = index + 1
                    # continue
                else:
                    self._logger.warning(f"Invalid pin format {node_str} for {self.name}")
                    # self._valid = False
                    # return 
                    continue
            if isinstance(node_str, int):
                internal_node = node_str
                name = f'{internal_node}'
            else:
                node_str = node_str.strip()
                i = node_str.find(' ')
                if i >= 1:
                    try:
                        internal_node = int(node_str[:i])
                    except ValueError:
                        self._logger.error(f"Invalid pin format {node_str} for {self.name}")
                    name = node_str[i:].strip()
            if internal_node and name:
                _ = Pin(index, name, internal_node)
                self._nodes.append(_)
                # Fixme: useful ?
                # self._pin_map[internal_node] = _
                self._pin_map[name] = _
            else:
                self._valid = False

    ##############################################

    def __repr__(self) -> str:
        return f'Subcircuit {self._name} {self._nodes} "{self._description}"'

    def __bool__(self) -> bool:
        return self._valid

    def __len__(self) -> int:
        return len(self._nodes)

    def __iter__(self) -> Iterator[str]:
        return iter(self._nodes)

    @property
    def number_of_pins(self) -> int:
        # Fixme: self._nodes ?
        return len(self)

    ##############################################

    @property
    def pin_names(self) -> list[str]:
        return list(self._pin_map.keys())

    ##############################################

    def to_yaml(self) -> dict:
        _ = super().to_yaml()
        _.update({'nodes': [{'index': _.index, 'name': _.name, 'internal_node': _.internal_node} for _ in self._nodes]})
        return _

    ##############################################

    def map_nodes(self, **kwargs) -> list[str]:
        # Fixme: if name is not valid python id ?
        N = self.number_of_pins
        if len(kwargs) != N:
            raise NameError(f"wrong number of nodes {len(kwargs)} != {N}")
        nodes = [None] * N
        for name, node in kwargs.items():
            pin = self._pin_map[name]
            nodes[pin.index] = node
        return nodes

####################################################################################################

class SpiceInclude:

    INCLUDE_PREFIX = '.include '

    _logger = _module_logger.getChild('SpiceInclude')

    ##############################################

    def __init__(self, path: str | Path, rewrite_yaml: bool = False, recurse: bool = False, section: str | None = None) -> None:
        self._path = Path(path)   # .resolve()
        self._extension = None

        self._description = ''
        self._inner_includes = []
        self._inner_libraries = []
        self._models = {}
        self._subcircuits = {}
        self._digest = None
        self._recursive_digest = None
        self._recurse = recurse
        self._section = section

        # Fixme: check still valid !
        if not rewrite_yaml and self.has_yaml:
            self.load_yaml()
            # self.dump()
        else:
            self.parse()
            self.write_yaml()
            if self._recurse:
                self._process_inner_includes()

    ##############################################

    def _process_inner_includes(self) -> None:
        """Process inner includes recursively and add their models and subcircuits.
        
        This method processes all inner includes detected during parsing and 
        adds their models and subcircuits to this SpiceInclude instance.
        """
        processed_paths = set()  # Track processed paths to avoid circular includes
        processed_paths.add(str(self._path.resolve()))
        
        # Create a list to store inner include instances
        includes = []
        
        # Process each inner include path
        for path_str in self._inner_includes:
            try:
                path = Path(path_str)
                # Ensure path is absolute by resolving it relative to parent directory if needed
                if not path.is_absolute():
                    path = (self._path.parent / path).resolve()
                else:
                    path = path.resolve()
                
                # Skip if we've already processed this path
                if str(path) in processed_paths:
                    self._logger.info(f"Skipping already processed include: {path}")
                    continue
                
                # Check if file exists
                if not path.exists():
                    self._logger.warning(f"Include file not found: {path}")
                    continue
                
                self._logger.info(f"Processing inner include {path}")
                include = SpiceInclude(path, recurse=self._recurse)
                includes.append(include)
                processed_paths.add(str(path))
                
                # Add models from this include
                for model in include.models:
                    if model.name in self._models:
                        self._logger.warning(f"Duplicate model {model.name} from {path}, keeping original")
                    else:
                        self._models[model.name] = model
                
                # Add subcircuits from this include
                for subcircuit in include.subcircuits:
                    if subcircuit.name in self._subcircuits:
                        self._logger.warning(f"Duplicate subcircuit {subcircuit.name} from {path}, keeping original")
                    else:
                        self._subcircuits[subcircuit.name] = subcircuit
                
            except Exception as e:
                self._logger.error(f"Failed to process inner include {path_str}: {str(e)}")
        
        # Replace the path strings with actual SpiceInclude instances 
        # for recursive digest computation
        self._inner_includes = includes

    # def dump(self) -> None:
    #     print(self._path)
    #     for _ in self._models:
    #         print(repr(_))
    #     for _ in self._subcircuits:
    #         print(repr(_))

    ##############################################

    @property
    def path(self) -> Path:
        return self._path

    @property
    def extension(self) -> str:
        # Fixme: cache ???
        if self._extension is None:
            self._extension = self._path.suffix.lower()
        return self._extension

    @property
    def mtime(self) -> datetime:
        _ = self._path.stat().st_mtime
        return datetime.fromtimestamp(_)

    @property
    def yaml_path(self) -> str:
        # self._path.parent(f'{self._path.name}{YAML_EXTENSION}')
        return self._path.with_suffix(YAML_EXTENSION)

    @property
    def has_yaml(self) -> bool:
        return self.yaml_path.exists()

    @property
    def description(self) -> str:
        return self._description

    ##############################################

    @property
    def inner_includes(self) -> Iterator[Path]:
        return iter(self._inner_includes)

    @property
    def inner_libraries(self) -> Iterator[Path]:
        return iter(self._inner_libraries)

    ##############################################

    @property
    def subcircuits(self) -> Iterator[Subcircuit]:
        # Fixme: iter ?
        return iter(self._subcircuits.values())

    @property
    def models(self) -> Iterator[Model]:
        return iter(self._models.values())

    def __getitem__(self, name: str) -> Subcircuit | Model:
        if name in self._subcircuits:
            return self._subcircuits[name]
        elif name in self._models:
            return self._models[name]
        else:
            message = f"Library {self.path} contains:{NEWLINE}"
            def add_line(item):
                nonlocal message
                message += f'    - {item}{NEWLINE}'
            message += f"  Subcircuits:{NEWLINE}"
            for _ in self._subcircuits:
                add_line(_)
            for _ in self._models:
                add_line(_)
            self._logger.error(message)
            raise KeyError(name)

    ##############################################

    def parse(self) -> None:
        self._logger.info(f"Parse {self._path}")
        try:
            spice_file = SpiceFile(self._path, self._section)
        except ParseError as exception:
            # Parse problem with this file, so skip it and keep going.
            self._logger.warn(f"Parse error in Spice library {self._path}{NEWLINE}{exception}")
        
        # Convert include paths to absolute paths if they are relative
        self._inner_includes = []
        for include_path in spice_file.includes:
            path = Path(str(include_path))
            # If path is not absolute, make it absolute using self.path's parent directory
            if not path.is_absolute():
                path = self._path.parent / path
            self._inner_includes.append(path)
        
        # Process library files separately
        # self._inner_libraries = []
        # for lib in spice_file.libraries:
        #     path = Path(str(lib.path))
        #     # If path is not absolute, make it absolute using self.path's parent directory
        #     if not path.is_absolute():
        #         path = self._path.parent / path
        #     self._inner_libraries.append(path)
        
        for subcircuit in spice_file.subcircuits:
            # name = self._suffix_name(subcircuit.name)
            _ = Subcircuit(self, subcircuit.name, subcircuit.nodes)
            self._subcircuits[_.name] = _
        if spice_file.is_only_model:
            for model in spice_file.models:
                # name = self._suffix_name(model.name)
                _ = Model(self, model.name, model.type)
                self._models[_.name] = _

    ##############################################

    def _suffix_name(self, name: str) -> str:
        if self.extension.endswith('@xyce'):
            name += '@xyce'
        return name

    ##############################################

    def write_yaml(self):
        with open(self.yaml_path, 'w', encoding='utf8') as fh:
            data = {
                # 'path': str(self._path),
                'path': str(self.path),
                'date': self.mtime.isoformat(),
                'digest': self.digest,
                'description': self._description,
            }
            if self._models:
                data['models'] = [_.to_yaml() for _ in self.models]
            if self._subcircuits:
                data['subcircuits'] = [_.to_yaml() for _ in self.subcircuits]
            if self._inner_includes:
                data['inner_includes'] = [str(_) for _ in self._inner_includes]
            if self._inner_libraries:
                data['inner_libraries'] = self._inner_libraries
            # data['recursive_digest'] = self.recursive_digest
            fh.write(yaml.dump(data, sort_keys=False))

    ##############################################

    def _read_yaml(self) -> dict:
        with open(self.yaml_path, 'r', encoding='utf8') as fh:
            data = yaml.load(fh.read(), Loader=yaml.SafeLoader)
        return data

    ##############################################

    def load_yaml(self) -> None:
        self._logger.info(f"Load {self.yaml_path}")
        data = self._read_yaml()
        self._description = data['description']
        if 'models' in data:
            models = [Model.from_yaml(self, _) for _ in data['models']]
            self._models = {_.name: _ for _ in models}
        if 'subcircuits' in data:
            subcircuits = [Subcircuit.from_yaml(self, _) for _ in data['subcircuits']]
            self._subcircuits = {_.name: _ for _ in subcircuits}
        if 'inner_includes' in data:
            self._inner_includes = data['inner_includes']
        if 'inner_libraries' in data:
            self._inner_libraries = data['inner_libraries']

    ##############################################

    def _compute_digest(self, func) -> str:
        with open(self._path, 'rb') as fh:
            _ = func(fh.read()).hexdigest()
        return _

    ##############################################

    def _sha1(self) -> str:
        return self._compute_digest(hashlib.sha1)

    ##############################################

    def _compute_recursive_digest(self) -> str:
        if self._inner_includes:
            return self._digest + '[' + '/'.join([_.digest for _ in self._inner_includes]) + ']'
        else:
            return self._digest

    ##############################################

    @property
    def digest(self) -> str:
        if self._digest is None:
            self._digest = self._sha1()
        return self._digest

    @property
    def recursive_digest(self) -> str:
        if self._recursive_digest is None:
            self._recursive_digest = self._compute_recursive_digest()
        return self._recursive_digest




import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice import Circuit, SubCircuit, SubCircuitFactory
from InSpice.Spice.Parser.HighLevelParser import SpiceSource
from InSpice.Spice.Parser import Translator
from InSpice.Unit import *

raw_spice = '''
*.tile Qucs 25.1.1  ...
.INCLUDE "/usr/share/qucs-s/xspice_cmlib/include/ngspice_mathfunc.inc"

.SUBCKT SpiceOpamp_LM358 	gnd 1 2 3 4 5
*
C1   11 12 5.544E-12
C2    6  7 20.00E-12
DC    5 53 DX
DE   54  5 DX
DLP  90 91 DX
DLN  92 90 DX
DP    4  3 DX
EGND 99  0 POLY(2) (3,0) (4,0) 0 .5 .5
FB    7 99 POLY(5) VB VC VE VLP VLN 0 15.91E6 -20E6 20E6 20E6 -20E6
GA    6  0 11 12 125.7E-6
GCM   0  6 10 99 7.067E-9
IEE   3 10 DC 10.04E-6
HLIM 90  0 VLIM 1K
Q1   11  2 13 QX
Q2   12  1 14 QX
R2    6  9 100.0E3
RC1   4 11 7.957E3
RC2   4 12 7.957E3
RE1  13 10 2.773E3
RE2  14 10 2.773E3
REE  10 99 19.92E6
RO1   8  5 50
RO2   7 99 50
RP    3  4 30.31E3
VB    9  0 DC 0
VC 3 53 DC 2.100
VE   54  4 DC .6
VLIM  7  8 DC 0
VLP  91  0 DC 40
VLN   0 92 DC 40
.MODEL DX D(IS=800.0E-18)
.MODEL QX PNP(IS=800.0E-18 BF=250)
.ENDS
  
.SUBCKT Power_amp_arduino_subscheme Vout gnd Vin VCC 
QX2N2222A_1 VCC Vopamp Vout QMOD_X2N2222A_1 AREA=1
.MODEL QMOD_X2N2222A_1 npn (Is=14.34F Nf=1 Nr=1 Ikf=0.2847 Ikr=0 Vaf=74.03 Var=0 Ise=14.34F Ne=1.307 Isc=0 Nc=2 Bf=255.9 Br=6.092 Rbm=0 Irb=0 Rc=1 Re=0 Rb=10 Cje=22.01P Vje=0.75 Mje=0.377 Cjc=7.306P Vjc=0.75 Mjc=0.3416 Xcjc=1 Cjs=0 Vjs=0.75 Mjs=0 Fc=0.5 Tf=411.1P Xtf=3 Vtf=0 Itf=0.6 Tr=46.91N Kf=0 Af=1 Ptf=0 Xtb=1.5 Xti=3 Eg=1.11 Tnom=26.85 )
RG 0 _net2  100K
XOP1 0  Vin _net2 VCC 0 Vopamp SpiceOpamp_LM358
Rf _net2 Vout  1K 
RG1 Vin 0  100K
.ENDS
Rsh Vin _net0  1K
V2 VCC 0 DC 8
V4 Vin 0 DC 2
XSUB1 _net4 0 _net0 VCC Power_amp_arduino_subscheme
QX2N2222A_1 _net5 Vbase 0 QMOD_X2N2222A_1 AREA=1
.MODEL QMOD_X2N2222A_1 npn (Is=14.34F Nf=1 Nr=1 Ikf=0.2847 Ikr=0 Vaf=74.03 Var=0 Ise=14.34F Ne=1.307 Isc=0 Nc=2 Bf=255.9 Br=6.092 Rbm=0 Irb=0 Rc=1 Re=0 Rb=10 Cje=22.01P Vje=0.75 Mje=0.377 Cjc=7.306P Vjc=0.75 Mjc=0.3416 Xcjc=1 Cjs=0 Vjs=0.75 Mjs=0 Fc=0.5 Tf=411.1P Xtf=3 Vtf=0 Itf=0.6 Tr=46.91N Kf=0 Af=1 Ptf=0 Xtb=1.5 Xti=3 Eg=1.11 Tnom=26.85 )
VIb _net4 OutV DC 0
Rsh2 Vbase OutV  1K

'''
source = raw_spice

spice_source = SpiceSource(source=source, title_line=False)
for obj in spice_source.obj_lines:
    print(obj)
bootstrap_circuit = Translator.Builder().translate(spice_source)
bootstrap_source = str(bootstrap_circuit)
print(bootstrap_source)
# assert bootstrap_source == source
